import os
import time
from datetime import datetime

import pytest

from pykechain.enums import ServiceExecutionStatus, ServiceType, ServiceEnvironmentVersion
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError, APIError
# new in 1.13
from pykechain.models import Service
from pykechain.utils import temp_chdir
from tests.classes import TestBetamax


class TestServiceSetup(TestBetamax):
    """Only for test setup, will create a service with a debug script

    :ivar service: service with a debug.py script
    """

    def _create_service(self, name=None):
        """Creates a service with name, and adds a test_upload_script.py (debugging)"""
        # setUp
        new_service = self.project.create_service(
            name=name or 'Test upload script to service',
            description="Only used for testing - you can safely remove this",
            environment_version=ServiceEnvironmentVersion.PYTHON_3_8,
        )
        upload_path = os.path.join(self.test_assets_dir, 'tests', 'files', 'test_upload_script_to_service',
                                   'test_upload_script.py')

        # testing
        new_service.upload(pkg_path=upload_path)
        self.assertEqual(new_service._json_data['script_file_name'], 'test_upload_script.py')
        return new_service

    def setUp(self):
        super(TestServiceSetup, self).setUp()
        self.test_assets_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))
        self.service = self._create_service()

    def tearDown(self):
        self.service.delete()
        super(TestServiceSetup, self).tearDown()


class TestServices(TestBetamax):
    def _create_service(self, name=None):
        """Creates a service with name, and adds a test_upload_script.py (debugging)"""
        # setUp
        new_service = self.project.create_service(
            name=name or 'Test upload script to service',
            description="Only used for testing - you can safely remove this"
        )
        upload_path = os.path.join(self.test_assets_dir, 'tests', 'files', 'uploaded', 'test_upload_script.py')

        # testing
        new_service.upload(pkg_path=upload_path)
        new_service.refresh()
        self.assertEqual(new_service._json_data['script_file_name'], 'test_upload_script.py')
        return new_service

    def setUp(self):
        super(TestServices, self).setUp()
        self.test_assets_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))

    def test_retrieve_services(self):
        self.assertTrue(self.project.services())

    def test_retrieve_services_with_kwargs(self):
        # setUp
        retrieved_services_with_kwargs = self.project.services(script_type=ServiceType.PYTHON_SCRIPT)

        # testing
        self.assertTrue(retrieved_services_with_kwargs)
        for service in retrieved_services_with_kwargs:
            self.assertEqual(ServiceType.PYTHON_SCRIPT, service._json_data['script_type'])

    def test_retrieve_service_but_found_multiple(self):
        with self.assertRaises(MultipleFoundError):
            self.project.service(script_type=ServiceType.PYTHON_SCRIPT)

    def test_retrieve_single_service(self):
        services = self.project.services()
        self.assertTrue(services)
        service_1 = services[0]

        self.assertEqual(self.project.service(pk=service_1.id), service_1)

    def test_retrieve_service_by_name(self):
        service_name = 'Service Gears - Successful'
        service = self.project.service(name=service_name)
        self.assertTrue(service)
        self.assertEqual(service.name, service_name)

    def test_properties_of_service(self):
        service_name = 'Service Gears - Successful with Package'
        service = self.project.service(name=service_name)

        for key, value in service.__dict__.items():
            if str(key).startswith('_'):
                continue

            # Verified on is an optional variable
            if key == 'verified_on':
                continue

            with self.subTest(msg='{}: {}'.format(key, value)):
                self.assertIsNotNone(value)

    @pytest.mark.skipif("os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
                        reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided")
    def test_debug_service_execute(self):
        service_name = 'Service Gears - Successful'
        service = self.project.service(name=service_name)

        service_execution = service.execute()
        self.assertTrue(service_execution.status in ServiceExecutionStatus.values())
        if service_execution.status in (ServiceExecutionStatus.LOADING, ServiceExecutionStatus.RUNNING):
            # sleep 2000 ms
            time.sleep(2)
            service_execution.refresh()
            self.assertTrue(service_execution.status in ServiceExecutionStatus.values())

    @pytest.mark.skipif("os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
                        reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided")
    def test_service_context(self):
        some_activity = self.project.activities()[0]
        service = self.project.service(name='Service Gears - Successful')

        service_execution = service.execute(activity_id=some_activity.id)
        self.assertEqual(some_activity.id, service_execution.activity_id)

    def test_update_service(self):
        # setUp
        service_name = 'Service Gears - Successful'
        service = self.project.service(name=service_name)
        version_before = str(service.version)
        name_before = service_name
        name_after = 'Pykechain needs no debugging'
        description_before = str(service._json_data['description'])
        description_after = 'Pykechain is way too good for that'
        version_after = '-latest'

        # testing
        service.edit(name=name_after, description=description_after, version=version_after)
        service.refresh()
        self.assertEqual(service.name, name_after)
        self.assertEqual(service._json_data['description'], description_after)
        self.assertEqual(service.version, version_after)

        # tearDown
        service.edit(name=name_before, description=description_before, version=version_before)

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_service_clear_values(self):
        # setup
        initial_name = 'Service testing editing'
        initial_description = 'Description test'
        initial_version = '1.0'
        initial_run_as = 'kenode'
        initial_trusted = False
        initial_type = ServiceType.NOTEBOOK
        initial_env = ServiceEnvironmentVersion.PYTHON_3_8_NOTEBOOKS
        # TODO: to be removed in later versions of pykechain, only for temporal compatibility
        compatibility_env = "3.7_notebook"

        self.service = self.project.create_service(name=initial_name)

        self.service.edit(
            name=initial_name,
            description=initial_description,
            version=initial_version,
            type=initial_type,
            environment_version=initial_env,
            run_as=initial_run_as,
            trusted=initial_trusted
        )

        # Edit without mentioning values, everything should stay the same
        new_name = 'Changed service name'

        self.service.edit(
            name=new_name
        )

        # testing
        self.assertEqual(self.service.name, new_name)
        self.assertEqual(self.service.description, initial_description)
        self.assertEqual(self.service.version, initial_version)
        self.assertEqual(self.service.run_as, initial_run_as)
        self.assertEqual(self.service.type, initial_type)
        self.assertIn(self.service.environment, (initial_env, compatibility_env))
        self.assertEqual(self.service.trusted, initial_trusted)

        # Edit with clearing the values, name and status cannot be cleared
        self.service.edit(
            name=None,
            description=None,
            version=None,
            type=None,
            environment_version=None,
            run_as=None,
            trusted=None
        )

        self.assertEqual(self.service.name, new_name)
        self.assertEqual(self.service.description, str())
        self.assertEqual(self.service.version, str())
        self.assertEqual(self.service.type, initial_type)
        self.assertIn(self.service.environment, (initial_env, compatibility_env))
        self.assertEqual(self.service.run_as, initial_run_as)
        self.assertEqual(self.service.trusted, initial_trusted)

        # teardown
        self.service.delete()

    # test added in 3.1
    def test_retrieve_services_with_refs(self):
        # setup
        service_ref = 'service-gears-successful'
        service_name = 'Service Gears - Successful'
        service = self.project.service(ref=service_ref)

        # testing
        self.assertIsInstance(service, Service)
        self.assertTrue(service.name, service_name)


class TestServicesWithCustomUploadedService(TestServiceSetup):

    def test_update_service_incorrect_name(self):
        with self.assertRaises(IllegalArgumentError):
            self.service.edit(name=1234)

    def test_update_service_incorrect_description(self):
        with self.assertRaises(IllegalArgumentError):
            self.service.edit(description=True)

    def test_update_service_incorrect_version(self):
        with self.assertRaises(IllegalArgumentError):
            self.service.edit(version=['2.0'])

    def test_service_refresh_from_kechain(self):
        version_after = '-latest'
        self.service.edit(version=version_after)
        self.service.refresh()
        self.assertEqual(self.service.version, version_after)

    def test_get_executions_of_service(self):
        self.assertTrue(len(self.service.get_executions()) >= 0)

    def test_create_and_delete_service(self):
        service_name = 'Test service creation'
        new_service = self.project.create_service(service_name)

        self.assertTrue(new_service.name, service_name)
        self.assertTrue(new_service)

        # tearDown
        new_service.delete()
        with self.assertRaisesRegex(NotFoundError, 'fit criteria'):
            self.project.service(pk=new_service.id)

    def test_create_service_with_wrong_service_type(self):
        with self.assertRaisesRegex(IllegalArgumentError, 'must be an option from enum'):
            self.project.create_service(name='This service type does not exist', service_type='RUBY_SCRIPT')

    def test_create_service_with_wrong_environment_version(self):
        with self.assertRaisesRegex(IllegalArgumentError, 'must be an option from enum'):
            self.project.create_service(name='This env version does not exist', environment_version='0.0')

    def test_save_service_script(self):
        # setUp
        with temp_chdir() as target_dir:
            self.service.save_as(target_dir=target_dir)
            self.assertEqual(len(os.listdir(target_dir)), 1)

    def test_upload_script_to_service(self):
        # setUp
        upload_path = os.path.join(self.test_assets_dir, 'tests', 'files', 'test_upload_script_to_service',
                                   'test_upload_script.py')

        # testing
        self.service.upload(pkg_path=upload_path)
        # second upload modified filename
        self.assertRegex(self.service._json_data['script_file_name'], 'test_upload_\w+.py')

    def test_upload_script_to_service_with_wrong_path(self):
        # setUp
        upload_path = os.path.join(self.test_assets_dir, 'tests', 'files', 'uploaded', 'this_file_does_exists.not')

        # testing
        with self.assertRaisesRegex(OSError, 'Could not locate python package to upload in'):
            self.service.upload(pkg_path=upload_path)


# new in 1.13
class TestServiceExecutions(TestServiceSetup):
    def test_retrieve_service_executions(self):
        self.assertTrue(self.project.service_executions())

    def test_retrieve_service_executions_with_kwargs(self):
        # setUp
        limit = 15
        retrieved_executions_with_kwargs = self.project.service_executions(limit=limit)

        # testing
        self.assertTrue(len(retrieved_executions_with_kwargs) <= limit)

    def test_retrieve_single_service_execution(self):
        service_executions = self.project.service_executions()
        self.assertTrue(service_executions)
        service_execution_1 = service_executions[0]

        self.assertEqual(self.project.service_execution(pk=service_execution_1.id), service_execution_1)

    def test_retrieve_single_service_execution_but_found_none(self):
        with self.assertRaises(NotFoundError):
            self.project.service_execution(username='No service execution as this user does not exist')

    def test_retrieve_single_service_execution_but_found_multiple(self):
        # setUp
        service_execution = self.service.execute()
        while service_execution.status in [ServiceExecutionStatus.LOADING, ServiceExecutionStatus.RUNNING]:
            time.sleep(0.500)  # 200ms
            service_execution.refresh()

        self.service.execute()

        # testing
        with self.assertRaises(MultipleFoundError):
            self.project.service_execution(service=self.service.id)

    def test_service_execution_conflict(self):
        # setUp
        self.service.execute()

        # testing
        with self.assertRaisesRegex(APIError, "Conflict: Could not execute"):
            self.service.execute()

    def test_properties_of_service_execution(self):
        service_name = 'Service Gears - Successful'
        service = self.project.service(name=service_name)

        service_executions = self.project.service_executions(service=service.id, limit=1)
        self.assertTrue(service_executions)

        service_execution = service_executions[0]

        self.assertIsInstance(service_execution.service, Service)
        self.assertIsInstance(service_execution.started_at, datetime)
        self.assertIsInstance(service_execution.finished_at, datetime)

        for key, value in service_execution.__dict__.items():
            if str(key).startswith('_'):
                continue

            # Originating activity is an optional variable
            if key == 'activity_id':
                continue

            with self.subTest(msg='{}: {}'.format(key, value)):
                self.assertIsNotNone(value)

    @pytest.mark.skipif("os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
                        reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided")
    def test_debug_service_execution_terminate(self):
        service_execution = self.service.execute()
        self.assertEqual(service_execution.status, ServiceExecutionStatus.LOADING)
        time.sleep(2)
        service_execution.refresh()
        self.assertEqual(service_execution.status, ServiceExecutionStatus.RUNNING)
        service_execution.terminate()

        self.assertNotEqual(service_execution.status, ServiceExecutionStatus.FAILED,
                            "The service execution is status 'FAILED', please upload working debugging scripts before "
                            "running the tests")

    def test_log_of_service_execution(self):
        # setUp
        service_execution = self.service.execute()
        time.sleep(5)

        with temp_chdir() as target_dir:
            service_execution.get_log(target_dir=target_dir)
            log_file = os.path.join(target_dir, 'log.txt')
            self.assertTrue(log_file)



from typing import List, Any, Dict, Callable

from jsonschema import validate

from pykechain.enums import _AllRepresentations, PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.representations.representation_base import BaseRepresentation
from pykechain.models.validators.validator_schemas import representation_jsonschema_stub


class RepresentationsComponent(object):
    """
    Aggregate class to use representations on an object.

    To add representations to a Pykechain class, create an instance of this class in its __init__() method.

    .. versionadded:: 3.7
    """

    def __init__(
        self, parent_object, representation_options: Dict, update_method: Callable
    ):
        """
        Extract the json with the representation options.

        :param parent_object: Object to which this representation component is attached
        :param representation_options: json with list of representations
        :param update_method: method of the parent_object that is used to update the representations
        """
        self._parent_object = parent_object
        self._repr_options = representation_options
        self._update_method = update_method  # type: Callable

        # Construct representation objects
        self._representations = []  # type: List['AnyRepresentation']
        representations_json = self._repr_options
        for representation_json in representations_json:
            self._representations.append(
                BaseRepresentation.parse(
                    obj=self._parent_object, json=representation_json
                )
            )

    def get_representations(self) -> List["AnyRepresentation"]:
        """
        Get list of representation objects.

        :return: list of Representations
        :raises IllegalArgumentError if representations are set with incorrect options
        """
        return self._representations

    def set_representations(self, representations: List["AnyRepresentation"]) -> None:
        """Set the representations."""
        self._validate_representations(representations)

        # set the internal representation list
        self._representations = list(set(representations))

        # dump to _json options
        self._dump_representations()

        # update the options to KE-chain backend
        self._update_method(self._repr_options)

    def _validate_representations(self, representations: Any):
        """Check provided representation inputs."""
        if not isinstance(representations, (tuple, list)):
            raise IllegalArgumentError(
                "Should be a list or tuple with Representation objects, "
                "got {}".format(type(representations))
            )

        for r in representations:
            if not isinstance(r, BaseRepresentation):
                raise IllegalArgumentError(
                    "Representation '{}' should be a Representation object".format(r)
                )
            if not _valid_object_type(r, self._parent_object):
                raise IllegalArgumentError(
                    "Representation '{}' can not be added to '{}'.".format(
                        r, self._parent_object
                    )
                )
            r.validate_json()

    def _dump_representations(self):
        """Dump the representations as json inside the _repr_options dictionary."""
        representations_json = []
        for r in self._representations:
            json_format = r.as_json()
            validate(json_format, representation_jsonschema_stub)
            representations_json.append(json_format)

        self._repr_options = representations_json


def _valid_object_type(representation: BaseRepresentation, obj: "Base") -> bool:
    """
    Check whether the representation can be used on the provided object.

    :param representation: representation to check
    :type representation: BaseRepresentation
    :param obj: object to attach the representation to
    :type obj: Base
    :return: True if feasible, False if not.
    :rtype bool
    """
    rtype = representation.rtype
    if rtype == _AllRepresentations.CUSTOM_ICON:
        from pykechain.models import Activity, Scope

        return isinstance(obj, (Activity, Scope))
    else:
        from pykechain.models import Property

        if not isinstance(obj, Property):
            return False
        else:
            if rtype == _AllRepresentations.BUTTON:
                return obj.type in [
                    PropertyType.SINGLE_SELECT_VALUE,
                    PropertyType.MULTI_SELECT_VALUE,
                ]
            elif rtype == _AllRepresentations.DECIMAL_PLACES:
                return obj.type == PropertyType.FLOAT_VALUE
            elif rtype == _AllRepresentations.SIGNIFICANT_DIGITS:
                return obj.type == PropertyType.FLOAT_VALUE
            elif rtype == _AllRepresentations.THOUSANDS_SEPARATOR:
                return obj.type in [PropertyType.INT_VALUE, PropertyType.FLOAT_VALUE]
            elif rtype == _AllRepresentations.LINK_TARGET:
                return obj.type == PropertyType.LINK_VALUE
            elif rtype == _AllRepresentations.AUTOFILL:
                return obj.type in [
                    PropertyType.DATETIME_VALUE,
                    PropertyType.DATE_VALUE,
                    PropertyType.TIME_VALUE,
                ]
            else:
                return False



from unittest import TestCase

from pykechain.enums import PropertyType, FilterType, Multiplicity, ActivityRootNames
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import MultiReferenceProperty
from pykechain.models.base_reference import _ReferenceProperty
from pykechain.models.property_reference import (
    ActivityReferencesProperty,
    ScopeReferencesProperty,
    UserReferencesProperty,
)
from pykechain.models.validators import RequiredFieldValidator
from pykechain.models.value_filter import PropertyValueFilter, ScopeFilter
from pykechain.utils import find, is_uuid
from tests.classes import TestBetamax


class TestBaseReference(TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.base_ref = _ReferenceProperty(json=dict(), client=None)

    def test_prefilters(self):
        with self.assertRaises(NotImplementedError):
            self.base_ref.set_prefilters()

        with self.assertRaises(NotImplementedError):
            self.base_ref.get_prefilters()


class TestMultiReferenceProperty(TestBetamax):
    def setUp(self):
        super(TestMultiReferenceProperty, self).setUp()

        # reference Part target (model and 1 instance)
        _wheel_model = self.project.model(ref="wheel")

        props = [
            PropertyType.DATETIME_VALUE,
            PropertyType.SINGLE_SELECT_VALUE,
            PropertyType.BOOLEAN_VALUE,
            PropertyType.FLOAT_VALUE,
            PropertyType.INT_VALUE,
            PropertyType.CHAR_VALUE,
        ]

        self.target_model = self.project.create_model_with_properties(
            parent=_wheel_model.parent_id,
            name="__Wheel",
            multiplicity=Multiplicity.ONE_MANY,
            properties_fvalues=[dict(name=p, property_type=p) for p in props],
        )

        # reference property model (with a value pointing to a reference target part model
        self.part_model = self.project.model("Bike")

        self.datetime_prop = self.target_model.property(PropertyType.DATETIME_VALUE)
        self.ssl_prop = self.target_model.property(PropertyType.SINGLE_SELECT_VALUE)
        self.bool_prop = self.target_model.property(PropertyType.BOOLEAN_VALUE)
        self.float_prop = self.target_model.property(PropertyType.FLOAT_VALUE)
        self.integer_prop = self.target_model.property(PropertyType.INT_VALUE)
        self.char_prop = self.target_model.property(PropertyType.CHAR_VALUE)

        self.ref_prop_name = "__Test reference property"
        self.ref_prop_model = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model.id,
        )  # type: MultiReferenceProperty

        # reference property instance holding the value
        part_instance = self.part_model.instance()
        self.ref = find(
            part_instance.properties, lambda p: p.model_id == self.ref_prop_model.id
        )  # type: MultiReferenceProperty

    def tearDown(self):
        self.target_model.delete()
        self.ref_prop_model.delete()
        super(TestMultiReferenceProperty, self).tearDown()

    def test_referencing_a_model(self):
        # setUp
        wheel_model = self.project.model("Wheel")
        self.ref_prop_model.value = [wheel_model]

        # testing
        self.assertEqual(len(list(self.ref_prop_model.value)), 1)

    def test_referencing_multiple_instances_using_parts(self):
        # setUp
        wheel_model = self.project.model("Wheel")
        self.ref_prop_model.value = [wheel_model]
        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list

        # testing
        self.assertTrue(len(self.ref.value) >= 2)

    def test_referencing_multiple_instances_using_ids(self):
        # setUp
        wheel_model = self.project.model("Wheel")
        self.ref_prop_model.value = [wheel_model]
        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list
        self.ref._cached_values = None

        # testing
        self.assertTrue(len(self.ref.value) >= 2)

    def test_referencing_a_list_with_no_parts(self):
        # setUp
        fake_part = [15, 21, 26]

        # testing
        with self.assertRaises(ValueError):
            self.ref.value = fake_part

    def test_value_if_multi_ref_gives_back_all_parts(self):
        """because of #276 problem"""
        # setUp
        self.ref_prop_model.value = [self.target_model]

        wheel_instances = self.target_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list
        self.ref.refresh()

        # testing
        all_referred_parts = self.ref.value
        self.assertEqual(len(all_referred_parts), len(self.ref._value))

    def test_value_if_nothing_is_referenced(self):
        # setUp
        value_of_multi_ref = self.ref.value

        # testing
        self.assertFalse(self.ref.has_value())
        self.assertIsNone(value_of_multi_ref)

    def test_value_ids(self):
        wheel_instances = self.target_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]
        self.ref.value = wheel_instances_list
        self.ref.refresh()

        ids = self.ref.value_ids()

        self.assertIsInstance(ids, list)
        self.assertTrue(all(isinstance(v, str) for v in ids))

    def test_multi_ref_choices(self):
        # setUp
        self.project.part(ref="bike").add_with_properties(
            name='__Wheel 2',
            model=self.target_model,
            update_dict={
                PropertyType.BOOLEAN_VALUE: True,
            }
        )
        self.ref_prop_model.value = [self.target_model]
        possible_options = self.ref.choices()

        # testing
        self.assertEqual(2, len(possible_options))

        self.ref_prop_model.set_prefilters(
            property_models=[self.target_model.property(PropertyType.BOOLEAN_VALUE)],
            values=[True],
            filters_type=[FilterType.EXACT],
        )
        self.ref.refresh()

        possible_options = self.ref.choices()
        self.assertEqual(1, len(possible_options))

    def test_create_ref_property_referencing_part_in_list(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=[self.target_model],
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_referencing_id_in_list(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=[self.target_model.id],
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_wrongly_referencing_in_list(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.part_model.add_property(
                name=self.ref_prop_name,
                property_type=PropertyType.REFERENCES_VALUE,
                default_value=[12],
            )

    def test_create_ref_property_referencing_part(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model,
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_referencing_id(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model.id,
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_wrongly_referencing(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.part_model.add_property(
                name=self.ref_prop_name,
                property_type=PropertyType.REFERENCES_VALUE,
                default_value=True,
            )

    # new in 3.0
    def test_set_prefilters_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop  # decimal property
        spokes_property = self.integer_prop  # integer property
        rim_material_property = self.char_prop  # single line text

        self.ref_prop_model.set_prefilters(
            property_models=[
                diameter_property,
                spokes_property,
                rim_material_property,
                self.datetime_prop,
                self.ssl_prop,
                self.bool_prop,
            ],
            values=[30.5, 7, "Al", self.time, "Michelin", True],
            filters_type=[
                FilterType.GREATER_THAN_EQUAL,
                FilterType.LOWER_THAN_EQUAL,
                FilterType.CONTAINS,
                FilterType.GREATER_THAN_EQUAL,
                FilterType.CONTAINS,
                FilterType.EXACT,
            ],
        )
        self.assertIn("property_value", self.ref_prop_model._options["prefilters"])

        filter_string = self.ref_prop_model._options["prefilters"]["property_value"]
        filters = set(filter_string.split(","))

        # testing
        self.assertIn(
            "{}:{}:{}".format(
                diameter_property.id, 30.5, FilterType.GREATER_THAN_EQUAL
            ),
            filters,
        )
        self.assertIn(
            "{}:{}:{}".format(spokes_property.id, 7, FilterType.LOWER_THAN_EQUAL),
            filters,
        )
        self.assertIn(
            "{}:{}:{}".format(rim_material_property.id, "Al", FilterType.CONTAINS),
            filters,
        )
        self.assertIn(
            "{}:{}:{}".format(self.bool_prop.id, "true", FilterType.EXACT), filters
        )
        self.assertIn(
            "{}:{}:{}".format(self.ssl_prop.id, "Michelin", FilterType.CONTAINS),
            filters,
        )
        self.assertIn(
            "{}:{}:{}".format(
                self.datetime_prop.id, self.time, FilterType.GREATER_THAN_EQUAL
            ),
            filters,
        )

    def test_set_prefilters_with_tuples(self):
        prefilters_set = [
            PropertyValueFilter(self.float_prop, 15.3, FilterType.GREATER_THAN_EQUAL),
            PropertyValueFilter(self.char_prop, "Al", FilterType.CONTAINS),
        ]

        self.ref_prop_model.set_prefilters(prefilters=prefilters_set)

        prefilters_get = self.ref_prop_model.get_prefilters()

        self.assertTrue(prefilters_get, msg="No prefilters returned")
        first_filter = prefilters_get[0]

        self.assertEqual(self.float_prop.id, first_filter.id)
        self.assertEqual("15.3", first_filter.value)
        self.assertEqual(FilterType.GREATER_THAN_EQUAL, first_filter.type)

    def test_set_prefilters_with_validation(self):
        prefilter_good = PropertyValueFilter(self.float_prop, 15.3, FilterType.GREATER_THAN_EQUAL)
        prefilter_bad = PropertyValueFilter(self.part_model.property("Gears"), 15.3, FilterType.GREATER_THAN_EQUAL)

        # Validate automatically
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_good])
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(prefilters=[prefilter_bad])

        # Validate by providing the referenced model
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_good], validate=self.target_model)
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(prefilters=[prefilter_bad], validate=self.target_model)

        # Dont validate
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_good], validate=False)
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_bad], validate=None)

    def test_set_prefilters_on_reference_property_with_excluded_propmodels_and_validators(
        self,
    ):
        # The excluded propmodels and validators already set on the property should not be erased when
        # setting prefilters

        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(
            property_models=[diameter_property, spokes_property]
        )
        self.ref_prop_model.validators = [RequiredFieldValidator()]

        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL],
        )

        # testing Filters
        self.assertIn(
            "{}:{}:{}".format(
                diameter_property.id, 15.13, FilterType.GREATER_THAN_EQUAL
            ),
            self.ref_prop_model._options["prefilters"]["property_value"],
        )

        # testing Excluded props
        self.assertIn("propmodels_excl", self.ref_prop_model._options)
        excluded = self.ref_prop_model._options["propmodels_excl"]

        self.assertEqual(len(excluded), 2)
        self.assertIn(diameter_property.id, excluded)
        self.assertIn(spokes_property.id, excluded)

        # testing Validators
        self.assertTrue(self.ref_prop_model.validators)
        self.assertIsInstance(
            self.ref_prop_model._validators[0], RequiredFieldValidator
        )

    def test_clear_prefilters(self):
        # setUp
        prefilter_1 = PropertyValueFilter(
            property_model=self.float_prop,
            value=15.13,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_1])

        # Add filter for different property, see if the first one is removed (clear = True)
        prefilter_2 = PropertyValueFilter(
            property_model=self.integer_prop,
            value=13,
            filter_type=FilterType.EXACT,
        )
        self.ref_prop_model.set_prefilters(
            prefilters=[prefilter_2],
            clear=True,
        )

        self.ref_prop_model.refresh()
        live_prefilters = self.ref_prop_model.get_prefilters()

        # testing
        self.assertTrue(live_prefilters, msg="Expected at least 1 prefilter")
        self.assertEqual(1, len(live_prefilters), msg="Expected 1 prefilter")
        self.assertEqual(prefilter_2, live_prefilters[0])

    def test_overwrite_prefilters(self):
        # setUp
        prefilter_1 = PropertyValueFilter(
            property_model=self.float_prop,
            value=15.13,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )
        prefilter_2 = PropertyValueFilter(
            property_model=self.integer_prop,
            value=13,
            filter_type=FilterType.EXACT,
        )
        prefilter_3 = PropertyValueFilter(
            property_model=self.float_prop,
            value=15.13,
            filter_type=FilterType.EXACT,
        )
        # Set the initial filters
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_1, prefilter_2])

        # Add another filter for the same property to see if the first one is removed (overwrite = True)
        self.ref_prop_model.set_prefilters(prefilters=[prefilter_3], overwrite=True)

        self.ref_prop_model.refresh()
        live_prefilters = self.ref_prop_model.get_prefilters()

        # testing
        self.assertTrue(live_prefilters, msg="Expected at least 1 prefilter")
        self.assertEqual(2, len(live_prefilters), msg="Expected 2 prefilters")
        self.assertIn(prefilter_2, live_prefilters)
        self.assertIn(prefilter_3, live_prefilters)

    def test_set_prefilters_on_reference_property_using_uuid(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property.id, spokes_property.id],
            values=[30.5, 7],
            filters_type=[FilterType.GREATER_THAN_EQUAL, FilterType.LOWER_THAN_EQUAL],
        )

        # testing
        self.assertTrue("property_value" in self.ref_prop_model._options["prefilters"])
        self.assertTrue(
            "{}:{}:{}".format(diameter_property.id, 30.5, FilterType.GREATER_THAN_EQUAL)
            in self.ref_prop_model._options["prefilters"]["property_value"]
        )
        self.assertTrue(
            "{}:{}:{}".format(spokes_property.id, 7, FilterType.LOWER_THAN_EQUAL)
            in self.ref_prop_model._options["prefilters"]["property_value"]
        )

    def test_set_prefilters_on_reference_property_the_wrong_way(self):
        # setUp
        bike_gears_property = self.part_model.property(name="Gears")
        instance_diameter_property = self.target_model.instances()[0].property(
            PropertyType.FLOAT_VALUE
        )
        diameter_property = self.float_prop

        # testing
        # When prefilters are being set, but the property does not belong to the referenced part
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[bike_gears_property],
                values=[2],
                filters_type=[FilterType.GREATER_THAN_EQUAL],
            )

        # When prefilters are being set, but the property is an instance, not a model
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[instance_diameter_property],
                values=[3.33],
                filters_type=[FilterType.GREATER_THAN_EQUAL],
            )

        # When prefilters are being set, but the size of lists is not consistent
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[diameter_property],
                values=[3.33, 1.51],
                filters_type=[FilterType.GREATER_THAN_EQUAL, FilterType.CONTAINS],
            )

        # When prefilters are being set, but no UUIDs or `Property` objects are being used in property_models
        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.ref_prop_model.set_prefilters(
                property_models=[False, 301],
                values=[3.33, 1.51],
                filters_type=[FilterType.GREATER_THAN_EQUAL, FilterType.CONTAINS],
            )

    def test_set_excluded_propmodels_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(
            property_models=[diameter_property, spokes_property]
        )

        # testing
        self.assertEqual(len(self.ref_prop_model._options["propmodels_excl"]), 2)
        self.assertTrue(
            diameter_property.id in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            spokes_property.id in self.ref_prop_model._options["propmodels_excl"]
        )

    def test_set_excluded_propmodels_with_validation(self):
        excluded_model_good = self.float_prop
        excluded_model_bad = self.part_model.property("Gears")

        # Validate automatically
        self.ref_prop_model.set_excluded_propmodels(property_models=[excluded_model_good])
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(property_models=[excluded_model_bad])

        # Validate by providing the referenced model
        self.ref_prop_model.set_excluded_propmodels(property_models=[excluded_model_good], validate=self.target_model)
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(
                property_models=[excluded_model_bad], validate=self.target_model)

        # Dont validate
        self.ref_prop_model.set_excluded_propmodels(property_models=[excluded_model_good], validate=False)
        self.ref_prop_model.set_excluded_propmodels(property_models=[excluded_model_bad], validate=None)

    def test_set_excluded_propmodels_on_reference_property_using_uuid(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(
            property_models=[diameter_property.id, spokes_property.id]
        )

        # testing
        self.assertEqual(len(self.ref_prop_model._options["propmodels_excl"]), 2)
        self.assertTrue(
            diameter_property.id in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            spokes_property.id in self.ref_prop_model._options["propmodels_excl"]
        )

    def test_set_excluded_propmodels_on_reference_property_with_prefilters_and_validators(
        self,
    ):
        # The prefilters and validators already set on the property should not be erased when setting
        # excluded propmodels

        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop
        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL],
        )

        self.ref_prop_model.validators = [RequiredFieldValidator()]

        self.ref_prop_model.set_excluded_propmodels(
            property_models=[diameter_property, spokes_property]
        )

        # testing
        self.assertTrue("property_value" in self.ref_prop_model._options["prefilters"])
        self.assertTrue(
            "{}:{}:{}".format(
                diameter_property.id, 15.13, FilterType.GREATER_THAN_EQUAL
            )
            in self.ref_prop_model._options["prefilters"]["property_value"]
        )
        self.assertEqual(len(self.ref_prop_model._options["propmodels_excl"]), 2)
        self.assertTrue(
            diameter_property.id in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            spokes_property.id in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            isinstance(self.ref_prop_model._validators[0], RequiredFieldValidator)
        )

    def test_get_prefilters(self):
        # setUp
        self.ref_prop_model.set_prefilters(
            property_models=[self.float_prop],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL],
        )

        prefilters = self.ref_prop_model.get_prefilters()

        self.assertTrue(prefilters, msg="A prefilter should be set")

        self.assertIsInstance(prefilters, list, msg="By default, prefilters should be a list of tuples!")
        self.assertTrue(all(isinstance(pf, PropertyValueFilter) for pf in prefilters),
                        msg="Not every prefilter is a PropertyValueFilter object!")

        prefilters = self.ref_prop_model.get_prefilters(as_lists=True)

        self.assertIsInstance(prefilters, tuple, msg="Prefilters should be a tuple of 3 lists!")
        self.assertEqual(3, len(prefilters), msg="Expected 3 lists!")

        property_model_ids, values, filters = prefilters

        self.assertEqual(len(property_model_ids), len(values), len(filters))

    def test_set_excluded_propmodels_on_reference_property_the_wrong_way(self):
        # setUp
        bike_gears_property = self.part_model.property(name="Gears")
        instance_diameter_property = self.target_model.instances()[0].property(
            PropertyType.FLOAT_VALUE
        )

        # testing
        # When excluded propmodels are being set, but the property does not belong to the referenced part
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(
                property_models=[bike_gears_property]
            )

        # When excluded propmodels are being set, but the property is an instance, not a model
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(
                property_models=[instance_diameter_property]
            )

        # When excluded propmodels are being set, but no UUIDs or `Property` objects are being used in property_models
        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.ref_prop_model.set_excluded_propmodels(property_models=[False, 301])

    def test_add_excluded_propmodels_to_reference_property(self):
        # Set the initial excluded propmodels
        self.ref_prop_model.set_excluded_propmodels(property_models=[self.float_prop])

        # Add others, checks whether the initial ones are remembered (overwrite = False)
        self.ref_prop_model.set_excluded_propmodels(
            property_models=[self.integer_prop], overwrite=False
        )

        # testing
        self.assertEqual(len(self.ref_prop_model._options["propmodels_excl"]), 2)
        self.assertTrue(
            self.float_prop.id in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            self.integer_prop.id in self.ref_prop_model._options["propmodels_excl"]
        )

    def test_overwrite_excluded_propmodels_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        # Set the initial excluded propmodels
        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property])

        # Overwrite them
        self.ref_prop_model.set_excluded_propmodels(
            property_models=[spokes_property], overwrite=True
        )

        # testing
        self.assertEqual(len(self.ref_prop_model._options["propmodels_excl"]), 1)
        self.assertTrue(
            diameter_property.id not in self.ref_prop_model._options["propmodels_excl"]
        )
        self.assertTrue(
            spokes_property.id in self.ref_prop_model._options["propmodels_excl"]
        )

    def test_get_excluded_propmodel_ids(self):
        # setUp
        self.ref_prop_model.set_excluded_propmodels(property_models=[self.float_prop])

        excluded_propmodel_ids = self.ref_prop_model.get_excluded_propmodel_ids()

        self.assertTrue(excluded_propmodel_ids, msg="Excluded propmodels should be set")
        self.assertIsInstance(excluded_propmodel_ids, list)
        self.assertTrue(all(is_uuid(pk) for pk in excluded_propmodel_ids))

    def test_retrieve_scope_id(self):
        frame = self.project.part(name="Frame")

        ref_to_wheel = frame.property(name="Ref to wheel")

        self.assertEqual(ref_to_wheel.scope_id, self.project.id)

    def test_property_clear_selected_part_model(self):
        # setUp
        wheel_model = self.project.model("Wheel")
        self.ref_prop_model.value = [wheel_model]

        self.assertTrue(wheel_model.id == self.ref_prop_model.value[0].id)

        self.ref_prop_model.value = None

        # testing
        self.assertIsNone(self.ref_prop_model.value)

    def test_property_clear_referenced_part_instances(self):
        # setUp
        wheel_model = self.project.model("Wheel")
        self.ref_prop_model.value = [wheel_model]

        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list

        # clear referenced part instances
        self.ref.value = None

        # testing
        self.assertIsNone(self.ref.value)
        self.assertIsNotNone(self.ref_prop_model.value)

    def test_copy_reference_property_with_options(self):
        # setUp
        copied_ref_property = self.ref_prop_model.copy(
            target_part=self.part_model, name="__Copied ref property"
        )

        # testing
        self.assertEqual(copied_ref_property.name, "__Copied ref property")
        self.assertEqual(
            copied_ref_property.description, self.ref_prop_model.description
        )
        self.assertEqual(copied_ref_property.unit, self.ref_prop_model.unit)
        self.assertEqual(copied_ref_property.value, self.ref_prop_model.value)
        self.assertDictEqual(copied_ref_property._options, self.ref_prop_model._options)

        # tearDown
        copied_ref_property.delete()


class TestMultiReferencePropertyXScope(TestBetamax):
    def setUp(self):
        super(TestMultiReferencePropertyXScope, self).setUp()
        self.x_scope = self.client.create_scope(
            name="Cross_reference scope", tags=["x-scope-target"]
        )

        self.part_model = self.project.model("Bike")

        # Create reference property and retrieve its instance
        prop_name = "cross-scope reference property"
        self.x_reference_model = self.part_model.add_property(
            name=prop_name, property_type=PropertyType.REFERENCES_VALUE,
        )
        self.x_reference = self.part_model.instance().property(prop_name)

        # Define target model and create an instance
        product_root = self.x_scope.model("Product")
        self.x_target_model = product_root.add_model(
            name="Reference target", multiplicity=Multiplicity.ZERO_MANY
        )
        self.x_target = product_root.instance().add(
            model=self.x_target_model, name="Target instance"
        )

    def tearDown(self):
        self.x_reference_model.delete()
        self.x_scope.delete()
        super(TestMultiReferencePropertyXScope, self).tearDown()

    def test_set_model_value(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model]
        self.x_reference_model.refresh()

        # testing
        self.assertEqual(self.x_scope.id, self.x_reference_model.value[0].scope_id)
        self.assertEqual(self.x_target_model.id, self.x_reference_model.value[0].id)

    def test_set_model_value_using_id(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model.id]
        self.x_reference_model.refresh()

        # testing
        self.assertEqual(self.x_scope.id, self.x_reference_model.value[0].scope_id)
        self.assertEqual(self.x_target_model.id, self.x_reference_model.value[0].id)

    def test_set_value(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model.id]
        self.x_reference.refresh()
        self.x_reference.value = [self.x_target]

        self.assertTrue(len(self.x_reference.value) == 1)
        self.assertEqual(self.x_target.id, self.x_reference.value[0].id)


class TestActivityReference(TestBetamax):
    def setUp(self):
        super().setUp()
        root = self.project.model(name="Product")
        self.part = self.project.create_model(
            name="The part", parent=root, multiplicity=Multiplicity.ONE
        )
        self.prop = self.part.add_property(
            name="activity ref", property_type=PropertyType.ACTIVITY_REFERENCES_VALUE
        )

    def tearDown(self):
        if self.part:
            self.part.delete()
        super().tearDown()

    def test_create(self):
        self.assertIsInstance(self.prop, ActivityReferencesProperty)

    def test_value(self):
        wbs_root = self.project.activity(name=ActivityRootNames.WORKFLOW_ROOT)
        task = wbs_root.children()[0]

        self.prop.value = [task]
        self.prop.refresh()

        self.assertIsInstance(self.prop, ActivityReferencesProperty)
        self.assertIsNotNone(self.prop.value)
        self.assertEqual(task, self.prop.value[0])

    def test_value_ids(self):
        wbs_root = self.project.activity(name=ActivityRootNames.WORKFLOW_ROOT)
        task = wbs_root.children()[0]

        self.prop.value = [task]
        self.prop.refresh()

        ids = self.prop.value_ids()

        self.assertIsInstance(ids, list)
        self.assertTrue(all(isinstance(v, str) for v in ids))

    def test_reload(self):
        reloaded_prop = self.client.reload(obj=self.prop)

        self.assertFalse(
            self.prop is reloaded_prop,
            msg="Must be different Python objects, based on memory allocation",
        )
        self.assertEqual(
            self.prop,
            reloaded_prop,
            msg="Must be the same KE-chain prop, based on hashed UUID",
        )


class TestScopeReference(TestBetamax):
    def setUp(self):
        super().setUp()
        root = self.project.model(name="Product")

        self.part = self.project.create_model(
            name="Test part", parent=root, multiplicity=Multiplicity.ONE
        )
        self.scope_ref_prop = self.part.add_property(
            name="scope ref", property_type=PropertyType.SCOPE_REFERENCES_VALUE
        )  # type: ScopeReferencesProperty

    def tearDown(self):
        if self.part:
            self.part.delete()
        super().tearDown()

    def test_create(self):
        self.assertIsInstance(self.scope_ref_prop, ScopeReferencesProperty)

    def test_value(self):
        bike_project = self.client.scope(name="Bike Project")

        self.scope_ref_prop.value = [bike_project]

        self.assertIsInstance(self.scope_ref_prop, ScopeReferencesProperty)
        self.assertIsNotNone(self.scope_ref_prop.value)
        self.assertEqual(bike_project, self.scope_ref_prop.value[0])

    def test_value_ids(self):
        bike_project = self.client.scope(name="Bike Project")

        self.scope_ref_prop.value = [bike_project]
        self.scope_ref_prop.refresh()

        ids = self.scope_ref_prop.value_ids()

        self.assertIsInstance(ids, list)
        self.assertTrue(all(isinstance(v, str) for v in ids))

    def test_no_value(self):
        self.assertIsInstance(self.scope_ref_prop, ScopeReferencesProperty)
        self.assertIsNone(self.scope_ref_prop.value)

    def test_reload(self):
        reloaded_prop = self.client.reload(obj=self.scope_ref_prop)

        self.assertFalse(
            self.scope_ref_prop is reloaded_prop,
            msg="Must be different Python objects, based on memory allocation",
        )
        self.assertEqual(
            self.scope_ref_prop,
            reloaded_prop,
            msg="Must be the same KE-chain prop, based on hashed UUID",
        )

    def test_prefilters(self):
        live_filters = self.scope_ref_prop.get_prefilters()
        self.assertFalse(live_filters)

        tags = ["project", "catalog"]
        filters = [ScopeFilter(tag=tag) for tag in tags]

        self.scope_ref_prop.set_prefilters(
            prefilters=filters,
        )
        self.scope_ref_prop.refresh()
        live_filters = self.scope_ref_prop.get_prefilters()

        self.assertTrue(live_filters)
        self.assertIsInstance(live_filters, list)
        self.assertTrue(all(isinstance(pf, ScopeFilter) for pf in live_filters))

        self.scope_ref_prop.set_prefilters(clear=True)
        self.scope_ref_prop.refresh()
        live_filters = self.scope_ref_prop.get_prefilters()
        self.assertFalse(live_filters)

        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.scope_ref_prop.set_prefilters(prefilters=filters[0])


class TestUserReference(TestBetamax):
    def setUp(self):
        super().setUp()
        root = self.project.model(name="Product")

        self.part = self.project.create_model(
            name="Test part", parent=root, multiplicity=Multiplicity.ONE
        )
        self.user_ref_prop = self.part.add_property(
            name="user ref", property_type=PropertyType.USER_REFERENCES_VALUE
        )

    def tearDown(self):
        if self.part:
            self.part.delete()
        super().tearDown()

    def test_create(self):
        self.assertIsInstance(self.user_ref_prop, UserReferencesProperty)

    def test_value(self):
        user = self.client.user(name="User Test")

        self.user_ref_prop.value = [user]

        self.assertIsInstance(self.user_ref_prop, UserReferencesProperty)
        self.assertIsNotNone(self.user_ref_prop.value)
        self.assertEqual(user, self.user_ref_prop.value[0])

    def test_value_ids(self):
        user = self.client.user(name="User Test")

        self.user_ref_prop.value = [user]

        self.user_ref_prop.refresh()

        ids = self.user_ref_prop.value_ids()

        self.assertIsInstance(ids, list)
        self.assertTrue(all(isinstance(v, int) for v in ids))

    def test_no_value(self):
        self.assertIsInstance(self.user_ref_prop, UserReferencesProperty)
        self.assertIsNone(self.user_ref_prop.value)

    def test_reload(self):
        reloaded_prop = self.client.reload(obj=self.user_ref_prop)

        self.assertFalse(
            self.user_ref_prop is reloaded_prop,
            msg="Must be different Python objects, based on memory allocation",
        )
        self.assertEqual(
            self.user_ref_prop,
            reloaded_prop,
            msg="Must be the same KE-chain prop, based on hashed UUID",
        )



from typing import Dict, Optional, Text

from pykechain.utils import parse_datetime


class Base(object):
    """Base model connecting retrieved data to a KE-chain client.

    :ivar id: The UUID of the object (corresponds with the UUID in KE-chain).
    :type id: str
    :ivar name: The name of the object.
    :type name: basestring
    :ivar created_at: the datetime when the object was created if available (otherwise None)
    :type created_at: datetime or None
    :ivar updated_at: the datetime when the object was last updated if available (otherwise None)
    :type updated_at: datetime or None
    """

    def __init__(self, json: Dict, client: 'Client'):
        """Construct a model from provided json data."""
        self._json_data = json
        self._client = client  # type: 'Client'

        self.id = json.get('id', None)
        self.name = json.get('name', None)
        self.created_at = parse_datetime(json.get('created_at'))
        self.updated_at = parse_datetime(json.get('updated_at'))

    def __repr__(self):  # pragma: no cover
        return "<pyke {} '{}' id {}>".format(self.__class__.__name__, self.name, self.id[-8:])

    def __eq__(self, other):  # pragma: no cover
        if hasattr(self, 'id') and hasattr(other, 'id'):
            return self.id == other.id
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def refresh(self, json: Optional[Dict] = None, url: Optional[Text] = None, extra_params: Optional[Dict] = None):
        """Refresh the object in place.

        Can be called on the object without any arguments and should refresh the object inplace. If you want to
        use it in an advance way, you may call it with a json response from the server or provide the url to
        refetch the object from the server if the url cant be determined from the object itself.

        It is using the `Client.reload()` function to re-retrieve the object in a backend API call.

        :param json: (optional) json dictionary from a response from the server, will re-init object
        :type json: None or dict
        :param url: (optional) url to retrieve the object again, typically an identity url api/<service>/id
        :type url: None or basestring
        :param extra_params: (optional) additional paramenters (query params) for the request eg dict(fields='__all__')
        :type extra_params: None or dict
        """
        if json and isinstance(json, dict):
            self.__init__(json=json, client=self._client)
        else:
            src = self._client.reload(self, url=url, extra_params=extra_params)
            self.__dict__.update(src.__dict__)


class BaseInScope(Base):
    """
    Base model for KE-chain objects coupled to a scope.

    :ivar scope_id: UUID of the Scope
    :type scope_id: str
    """

    def __init__(self, json, *args, **kwargs):
        """Append the scope ID to the attributes of the base object."""
        super().__init__(json, *args, **kwargs)

        self.scope_id = json.get('scope_id', json.get('scope', None))
        self._scope = None  # type: Optional['Scope']

    @property
    def scope(self):
        """
        Scope this object belongs to.

        This property will return a `Scope` object. It will make an additional call to the KE-chain API.

        :return: the scope
        :type: :class:`pykechain.models.Scope`
        :raises NotFoundError: if the scope could not be found
        """
        if not self._scope:
            self._scope = self._client.scope(pk=self.scope_id, status=None)
        return self._scope



import datetime
import time
import warnings
from test.support import EnvironmentVarGuard
from unittest import TestCase, skipIf

import pytz

from pykechain.client import Client
from pykechain.enums import ScopeStatus
from pykechain.exceptions import ForbiddenError, ClientError, NotFoundError, IllegalArgumentError, APIError
from pykechain.models import Team, Base
from pykechain.models.scope import Scope
from tests.classes import TestBetamax
from tests.utils import TEST_FLAG_IS_WIM2


class TestClient(TestCase):
    def setUp(self):
        self.env = EnvironmentVarGuard()

    def test_init_default_url(self):
        client = Client()

        self.assertEqual(client.api_root, 'http://localhost:8000/')

    def test_init_custom_url(self):
        client = Client(url='http://testing.com:1234')

        self.assertEqual(client.api_root, 'http://testing.com:1234')

    def test_init_no_login(self):
        client = Client()

        self.assertNotIn('Authorization', client.headers)
        self.assertIsNone(client.auth)

    def test_init_basic_auth(self):
        client = Client()

        client.login('someuser', 'withpass')

        self.assertNotIn('Authorization', client.headers)
        self.assertTrue(client.auth)

    def test_init_token(self):
        client = Client()
        PSEUDO_TOKEN = '123123'

        client.login(token=PSEUDO_TOKEN)

        self.assertTrue(client.headers['Authorization'], 'Token {}'.format(PSEUDO_TOKEN))
        self.assertIsNone(client.auth)

    def test_init_no_ssl(self):
        client = Client(check_certificates=False)

        self.assertFalse(client.session.verify)

    # 1.12
    def test_client_raises_error_with_false_url(self):
        with self.assertRaises(ClientError):
            Client(url='wrongurl')

    def test_client_from_env(self):
        """
        Test if the client does not provide a warning when no env file can be found

            Userwarning: 'Could not any envfile.'
            '.../lib/python3.5/site-packages/envparse.py'
        """
        self.env.set('KECHAIN_URL', 'http://localhost:8000')
        with self.env:
            with warnings.catch_warnings(record=True) as captured_warnings:
                client = Client.from_env()
                self.assertEqual(len(captured_warnings), 0)

    # noinspection PyTypeChecker
    def test_reload(self):
        client = Client()

        not_a_kechain_object = 3
        with self.assertRaises(IllegalArgumentError, msg='Reload must receive an object of type Base.'):
            client.reload(not_a_kechain_object)

        empty_kechain_object = Base(json=dict(name='empty', id='1234567890'), client=client)
        with self.assertRaises(IllegalArgumentError, msg='Reload cant find API resource for type Base'):
            client.reload(empty_kechain_object)


class TestClientLive(TestBetamax):

    def setUp(self):
        super().setUp()
        self.temp_scope = None

    def tearDown(self):
        """Teardown the test, remove the scope.

        When you create a scope, assign it to self.temp_scope such that it is deleted when the test is complete
        """
        if self.temp_scope:
            self.temp_scope.delete()
        super().tearDown()

    def test_login(self):
        self.assertTrue(self.project.parts())

    def test_no_login(self):
        self.client.login('wrong', 'user')

        with self.assertRaises(ForbiddenError):
            self.client.parts()

    # 3.6.3
    def test_get_current_user(self):
        user = self.client.current_user()

        from pykechain.models import User
        self.assertIsInstance(user, User)

        self.client.auth = None
        self.client.headers = None
        with self.assertRaises(APIError):
            self.client.current_user()

    # 3.7.0
    def test_reload_deleted_object(self):
        bike_model = self.project.model(name='Bike')
        bike = bike_model.instance()
        wheel_model = bike_model.child(name='Wheel')

        wheel_name = '__new wheel: test for reloading'
        new_wheel_1 = bike.add(name=wheel_name, model=wheel_model)
        new_wheel_2 = self.project.part(name=wheel_name)

        self.assertFalse(new_wheel_1 is new_wheel_2, "Parts must be separate in Python memory")
        self.assertEqual(new_wheel_1, new_wheel_2, "Parts must be identical in UUIDs")

        new_wheel_2.delete()
        with self.assertRaises(NotFoundError):
            self.project.part(name=wheel_name)

        with self.assertRaises(NotFoundError, msg="Wheel cant be reloaded after it was deleted"):
            self.client.reload(new_wheel_1)

    # 2.6.0
    def test_create_scope(self):
        # setUp
        client = self.client
        scope_name = 'New scope (pykechain testing - remove me)'
        scope_description = 'This is a new scope for testing'
        scope_status = ScopeStatus.ACTIVE
        scope_tags = ['test_tag', 'new_project_tag']
        scope_start_date = datetime.datetime(2019, 4, 12, tzinfo=pytz.UTC)
        scope_due_date = datetime.datetime(2020, 4, 12, tzinfo=pytz.UTC)
        scope_team = client.teams()[0]

        self.temp_scope = client.create_scope(
            name=scope_name,
            description=scope_description,
            status=scope_status,
            tags=scope_tags,
            start_date=scope_start_date,
            due_date=scope_due_date,
            team=scope_team
        )

        # testing
        self.assertEqual(self.temp_scope.name, scope_name)
        self.assertEqual(scope_start_date, self.temp_scope.start_date)
        self.assertEqual(scope_due_date, self.temp_scope.due_date)
        self.assertIn(scope_tags[0], self.temp_scope.tags)
        self.assertIn(scope_tags[1], self.temp_scope.tags)
        self.assertEqual(scope_team, self.temp_scope.team)

    def test_create_scope_with_team_name(self):
        # setUp
        team_name = self.client.teams()[0].name

        self.temp_scope = self.client.create_scope(
            name='New scope using the team name',
            team=team_name
        )

        # testing
        self.assertTrue(self.temp_scope.team)
        self.assertIsInstance(self.temp_scope.team, Team)
        self.assertEqual(team_name, self.temp_scope.team.name)

    def test_create_scope_with_team_uuid(self):
        # setUp
        team_id = self.client.teams()[0].id

        self.temp_scope = self.client.create_scope(
            name='New scope using the team ID',
            team=team_id
        )

        # testing
        self.assertTrue(self.temp_scope.team)
        self.assertIsInstance(self.temp_scope.team, Team)
        self.assertEqual(team_id, self.temp_scope.team.id)

    def test_create_scope_no_arguments(self):
        # setUp
        self.temp_scope = self.client.create_scope(name='New scope no arguments')

        # testing
        self.assertEqual(self.temp_scope.name, 'New scope no arguments')
        self.assertTrue(self.temp_scope._json_data['start_date'])
        self.assertFalse(self.temp_scope.tags)

    # noinspection PyTypeChecker
    def test_create_scope_with_wrong_arguments(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name=12)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name='Failed scope', status='LIMBO')
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name='Failed scope', description=True)
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name='Failed scope', tags='One tag no list')
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name='Failed scope', tags=[12, 'this', 'fails'])
        with self.assertRaises(IllegalArgumentError):
            self.client.create_scope(name='Failed scope', team=['Fake team'])

    def test_clone_scope(self):
        # setUp
        self.temp_scope = self.client.clone_scope(source_scope=self.project)

        # testing
        self.assertIsInstance(self.temp_scope, Scope)
        self.assertEqual(self.temp_scope.name, 'CLONE - {}'.format(self.project.name))
        self.assertEqual(self.temp_scope.status, self.project.status)
        self.assertEqual(self.temp_scope.start_date, self.project.start_date)
        self.assertEqual(self.temp_scope.due_date, self.project.due_date)
        self.assertEqual(self.temp_scope.description, self.project.description)
        self.assertListEqual(self.temp_scope.tags, self.project.tags)

    def test_clone_scope_with_arguments(self):
        # setUp
        now = datetime.datetime.now()
        name = 'test clone'
        status = ScopeStatus.CLOSED
        description = 'test description'
        tags = ['one', 'two']
        team = self.project.team

        self.temp_scope = self.client.clone_scope(
            name=name,
            source_scope=self.project,
            status=status,
            start_date=now,
            due_date=now,
            description=description,
            tags=tags,
            team=team,
        )

        self.assertIsInstance(self.temp_scope, Scope)
        self.assertEqual(self.temp_scope.name, name)
        self.assertEqual(self.temp_scope.status, status)
        self.assertEqual(self.temp_scope.start_date, now)
        self.assertEqual(self.temp_scope.due_date, now)
        self.assertEqual(self.temp_scope.description, description)
        self.assertEqual(self.temp_scope.team, team)
        self.assertListEqual(self.temp_scope.tags, tags)

    def test_scope_delete(self):
        new_scope = self.project.clone(asynchronous=False)
        self.assertNotEqual(self.project.id, new_scope.id)
        new_scope.delete(asynchronous=False)

        with self.assertRaisesRegex(NotFoundError, 'fit criteria'):
            # throw in arbitrary sleep to give backend time to actually delete the scope.
            time.sleep(1)
            self.client.scope(pk=new_scope.id)


class TestCloneScopeAsync(TestBetamax):

    def setUp(self):
        super().setUp()
        self.temp_scope = None
        self.source = self.client.create_scope(
            name='_Async to be cloned scope SOURCE',
        )
        time.sleep(3)

    def tearDown(self):
        self.source.delete()
        if self.temp_scope:
            self.temp_scope.delete()
        super().tearDown()

    def test_clone_asynchronous(self):
        # setUp
        clone_name = '_Async cloned scope TARGET'

        self.client.clone_scope(
            name=clone_name,
            source_scope=self.source,
            asynchronous=True,
        )

        for _ in range(5):
            try:
                self.temp_scope = self.client.scope(name=clone_name)
            except NotFoundError:
                time.sleep(3)
                continue
            else:
                break

        # testing
        self.assertIsInstance(self.temp_scope, Scope)


@skipIf(not TEST_FLAG_IS_WIM2, reason="This tests is designed for WIM version 2, expected to fail on older WIM")
class TestClientAppVersions(TestBetamax):
    def test_retrieve_versions(self):
        """Test to retrieve the app versions from KE-chain"""
        app_versions = self.project._client.app_versions
        self.assertTrue(isinstance(app_versions, list))
        self.assertTrue(isinstance(app_versions[0], dict))
        self.assertTrue(set(app_versions[0].keys()),
                        {'app', 'label', 'version', 'major', 'minor', 'patch', 'prerelease'})

    def test_compare_versions(self):
        """Multitest to check all the matchings versions"""

        self.assertTrue(self.client.match_app_version(app='kechain2.core.wim', version='>=1.0.0'))
        self.assertTrue(self.client.match_app_version(label='wim', version='>=1.0.0'))
        self.assertFalse(self.client.match_app_version(label='wim', version='==0.0.1'))

        # value error
        # wrong version string (no semver string) to check against
        with self.assertRaises(ValueError):
            self.client.match_app_version(app='kechain2.core.wim', version='1.0')

        # right version, no operand in version
        with self.assertRaises(ValueError):
            self.client.match_app_version(app='kechain2.core.wim', version='1.0.0')

        # wrong operand (should be ==)
        with self.assertRaises(ValueError):
            self.client.match_app_version(app='kechain2.core.wim', version='=1.0.0')

        # not found a match version
        with self.assertRaises(IllegalArgumentError):
            self.client.match_app_version(app='kechain2.core.wim', version='')

        # no version found on the app kechain2.metrics
        with self.assertRaises(NotFoundError):
            self.client.match_app_version(app='kechain2.core.activitylog', version='>0.0.0', default=None)

        # no version found on the app kechain2.metrics, default = True, returns True
        self.assertTrue(self.client.match_app_version(app='kechain2.core.activitylog', version='>0.0.0', default=True))

        # no version found on the app kechain2.metrics, default = False, returns False
        self.assertFalse(self.client.match_app_version(app='kechain2.core.activitylog', version='>0.0.0', default=False))

        # no version found on the app kechain2.metrics, default = False, returns False
        # default is set to return False in the method
        self.assertFalse(self.client.match_app_version(app='kechain2.core.activitylog', version='>0.0.0'))

        # did not find the app
        with self.assertRaises(NotFoundError):
            self.client.match_app_version(app='nonexistingapp', version='>0.0.0', default=None)

        # did not find the app, but the default returns a False
        self.assertFalse(self.client.match_app_version(app='nonexistingapp', version='>0.0.0', default=False))

        # did not find the app, but the default returns a False, without providing the default, as the default is False
        self.assertFalse(self.client.match_app_version(app='nonexistingapp', version='>0.0.0'))



from pykechain.exceptions import _DeprecationMixin


class Property2(_DeprecationMixin):
    """A virtual object representing a KE-chain property.

    .. versionadded: 3.0
       This is a `Property` to communicate with a KE-chain 3 backend.

    :cvar bulk_update: flag to postpone update of properties until manually requested
    :type bulk_update: bool
    :ivar type: The property type of the property. One of the types described in :class:`pykechain.enums.PropertyType`
    :type type: str
    :ivar category: The category of the property, either `Category.MODEL` of `Category.INSTANCE`
    :type category: str
    :ivar description: description of the property
    :type description: str or None
    :ivar unit: unit of measure of the property
    :type unit: str or None
    :ivar model: the id of the model (not the model object)
    :type model: str
    :ivar output: a boolean if the value is configured as an output (in an activity)
    :type output: bool
    :ivar part: The (parent) part in which this property is available
    :type part: :class:`Part2`
    :ivar value: the property value, can be set as well as property
    :type value: Any
    :ivar validators: the list of validators that are available in the property
    :type validators: List[PropertyValidator]
    :ivar is_valid: if the property conforms to the validators
    :type is_valid: bool
    :ivar is_invalid: if the property does not conform to the validator
    :type is_invalid: bool
    """

    pass



import warnings
from abc import abstractmethod
from typing import Text, Union, Any, Dict, List, Optional

from pykechain.enums import FilterType, Category, PropertyType
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models.input_checks import check_base, check_enum, check_type, check_text


class BaseFilter(object):
    """Base class for any filters used in pykechain."""

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.format() == other.format()
        else:
            return False

    @abstractmethod
    def format(self) -> Text:  # pragma: no cover
        """Format a filter as a string."""
        pass

    @classmethod
    @abstractmethod
    def parse_options(cls, options: Dict) -> List['BaseFilter']:  # pragma: no cover
        """
        Convert the dict & string-based definition of a filter to a list of Filter objects.

        :param options: options dict from a property or meta dict from a widget.
        :return: list of Filter objects
        :rtype list
        """
        pass


class PropertyValueFilter(BaseFilter):
    """
    Property value filter, used for Part reference properties and filtered grid widgets.

    :ivar id: property model UUID
    :ivar value: value of the filter
    :ivar type: filter type
    """

    def __init__(
        self,
        property_model: Union[Text, 'Property'],
        value: Any,
        filter_type: FilterType,
    ):
        """Create PropertyValueFilter instance."""
        from pykechain.models import Property

        property_model_id = check_base(property_model, Property, "property_model")
        check_enum(filter_type, FilterType, "filter_type")

        self.id = property_model_id
        self.value = value
        self.type = filter_type

    def __repr__(self):
        return "PropertyValueFilter {}: {} ({})".format(self.type, self.value, self.id)

    def format(self) -> Text:
        """Format PropertyValueFilter as a string."""
        value = str(self.value).lower() if isinstance(self.value, bool) else self.value
        return "{}:{}:{}".format(self.id, value, self.type)

    def validate(self, part_model: 'Part') -> None:
        """
        Validate data of the PropertyValueFilter.

        :param part_model: Part model to which the filter will be applied.
        :returns None
        """
        from pykechain.models import Part

        check_base(part_model, Part, "part_model")

        try:
            prop = part_model.property(self.id)
        except NotFoundError:
            raise IllegalArgumentError(
                "Property value filters can only be set on properties belonging to the selected Part model.")

        if prop.category != Category.MODEL:
            raise IllegalArgumentError(
                'Property value filters can only be set on Property models, received "{}".'.format(prop))
        else:
            if prop.type == PropertyType.BOOLEAN_VALUE and self.type != FilterType.EXACT:
                warnings.warn("A PropertyValueFilter on a boolean property should use filter type `{}`".format(
                    FilterType.EXACT), Warning)

    @classmethod
    def parse_options(cls, options: Dict) -> List['PropertyValueFilter']:
        """
        Convert the dict & string-based definition of a property value filter to a list of PropertyValueFilter objects.

        :param options: options dict from a multi-reference property or meta dict from a filtered grid widget.
        :return: list of PropertyValueFilter objects
        :rtype list
        """
        check_type(options, dict, "options")

        prefilter_string = options.get("prefilters", {}).get("property_value")
        prefilter_string_list = prefilter_string.split(",") if prefilter_string else []

        prefilters = list()
        for pf_string in prefilter_string_list:
            prefilter_raw = pf_string.split(":")

            if len(prefilter_raw) == 1:   # FIXME encoding problem KE-chain
                prefilter_raw = pf_string.split("%3A")

            prefilters.append(PropertyValueFilter(*prefilter_raw))

        return prefilters


class ScopeFilter(BaseFilter):
    """
    Scope filter, used on scope reference properties.

    :ivar tag: string
    """

    def __init__(
            self,
            tag: Optional[Text] = None,
    ):
        """Create a ScopeFilter object."""
        self.tag = check_text(tag, "tag")

    def __repr__(self):
        return "ScopeFilter: tag: {}".format(self.tag)

    def format(self) -> Text:
        """Format ScopeFilter as a string."""
        return self.tag if self.tag else ""

    @classmethod
    def parse_options(cls, options: Dict) -> List['ScopeFilter']:
        """
        Convert the dict & string-based definition of a scope filter to a list of ScopeFilter objects.

        :param options: options dict from a scope reference property or meta dict from a scopes widget.
        :return: list of ScopeFilter objects
        :rtype list
        """
        check_type(options, dict, "options")

        filters_dict = options.get("prefilters", {})
        tags_string = filters_dict.get("tags__contains", "")
        tags = tags_string.split(",") if tags_string else []

        return [ScopeFilter(tag=tag) for tag in tags]



from unittest import TestCase

from pykechain.enums import FilterType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import PropertyValueFilter
from pykechain.models.value_filter import ScopeFilter
from tests.classes import TestBetamax


class TestPropertyValueFilter(TestBetamax):

    def setUp(self):
        super().setUp()

        self.bike = self.project.product_root_model("Bike")

        self.filter = PropertyValueFilter(
            property_model=self.bike.property(name="Gears"),
            value=15,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )

    # noinspection PyTypeChecker
    def test_creation(self):
        with self.assertRaises(IllegalArgumentError):
            PropertyValueFilter(property_model=self.bike, value=15, filter_type=FilterType.EXACT)

        with self.assertRaises(IllegalArgumentError):
            PropertyValueFilter(property_model=self.bike.property(name="Gears"), value=15, filter_type="this one plz")

    def test_repr(self):
        representation = self.filter.__repr__()

        self.assertIsInstance(representation, str)

    def test_format(self):
        string = self.filter.format()

        self.assertIsInstance(string, str)
        self.assertIn(":", string)

    def test_validate(self):
        self.filter.validate(part_model=self.bike)

        with self.assertRaises(IllegalArgumentError):
            self.filter.validate(part_model="Not a part model")

        bike_instance = self.bike.instance()
        property_instance = bike_instance.property(name="Gears")
        invalid_filter = PropertyValueFilter(
            property_model=property_instance,
            value=15,
            filter_type=FilterType.GREATER_THAN_EQUAL,
        )

        with self.assertRaises(IllegalArgumentError, msg="Property instance is not allowed"):
            invalid_filter.validate(part_model=bike_instance)

        with self.assertRaises(IllegalArgumentError, msg="Property instance should not match the bike model"):
            invalid_filter.validate(part_model=self.bike)

        boolean = self.bike.property(name="Sale?")
        boolean_filter = PropertyValueFilter(
            property_model=boolean,
            value=False,
            filter_type=FilterType.CONTAINS,
        )

        with self.assertWarns(Warning, msg="Using any filter but EXACT should be warned against"):
            boolean_filter.validate(part_model=self.bike)

        self.assertIn("false", boolean_filter.format(), msg="Boolean value should have been parsed to a string")

    def test_parse_options(self):
        for options in [
            dict(),
            dict(prefilters=dict()),
            dict(prefilters=dict(property_value="")),
        ]:
            with self.subTest():
                prefilters = PropertyValueFilter.parse_options(options=options)

                self.assertIsInstance(prefilters, list)
                self.assertEqual(0, len(prefilters))

        prefilters = PropertyValueFilter.parse_options(
            dict(prefilters=dict(property_value=self.filter.format()))
        )

        self.assertIsInstance(prefilters, list)
        self.assertEqual(1, len(prefilters))

    def test__eq__(self):
        second_filter = PropertyValueFilter(self.bike.property("Gears"), 15, FilterType.GREATER_THAN_EQUAL)
        third_filter = PropertyValueFilter(self.bike.property("Gears"), 16, FilterType.GREATER_THAN_EQUAL)
        fourth_filter = PropertyValueFilter(self.bike.property("Total height"), 15, FilterType.GREATER_THAN_EQUAL)
        fifth_filter = PropertyValueFilter(self.bike.property("Gears"), 15, FilterType.LOWER_THAN_EQUAL)

        self.assertEqual(self.filter, second_filter)
        self.assertNotEqual(self.filter, "Coffee filter")
        self.assertNotEqual(self.filter, third_filter)
        self.assertNotEqual(self.filter, fourth_filter)
        self.assertNotEqual(self.filter, fifth_filter)


class TestScopeFilter(TestCase):

    def setUp(self) -> None:
        self.filter = ScopeFilter(tag="Calculation")

    def test__repr__(self):
        representation = self.filter.__repr__()

        self.assertIsInstance(representation, str)

    def test__eq__(self):
        second_filter = ScopeFilter("Calculation")
        third_filter = ScopeFilter("Not a calculation")

        self.assertEqual(self.filter, second_filter)
        self.assertNotEqual(self.filter, third_filter)

    def test_format(self):
        string = self.filter.format()

        self.assertIsInstance(string, str)



from pykechain.enums import LanguageCodes
from pykechain.exceptions import NotFoundError, MultipleFoundError
from tests.classes import TestBetamax


class TestUsers(TestBetamax):

    def test_retrieve_users(self):
        self.assertTrue(self.client.users())

    def test_retrieve_single_unknown_user(self):
        with self.assertRaises(NotFoundError):
            self.client.user('123lladadwd')

    def test_user_attributes(self):
        attributes = ['_client', '_json_data', 'id', 'name', 'created_at', 'updated_at',
                      'username', 'timezone', 'language', 'email']

        obj = self.client.users()[0]
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(hasattr(obj, attribute),
                                "Could not find '{}' in the object: '{}'".format(attribute, obj.__dict__))

    def test_retrieve_single_multiple_user_raises_error(self):
        with self.assertRaises(MultipleFoundError):
            self.client.user()

    def test_retrieve_single_user_with_known_id(self):
        self.assertTrue(self.client.user(pk=1))

    def test_retrieve_single_user_with_known_username(self):
        self.assertTrue(self.client.user(username='testuser'))

    def test_retrieve_user_timezone(self):
        user_retrieved = self.client.user(username='testuser')
        timezone_retrieved = user_retrieved.timezone
        self.assertIn(timezone_retrieved.zone, ['Atlantic/Cape_Verde', 'UTC'])

    def test_retrieve_user_name(self):
        user_retrieved = self.client.user(username='testuser')
        name_retrieved = user_retrieved.name
        self.assertIn(name_retrieved, ['User Test', ''])

    def test_retrieve_default_name(self):
        user_retrieved = self.client.user(username='testuser')
        default_name = user_retrieved.default_name
        self.assertIsInstance(default_name, str)
        self.assertIn(default_name, user_retrieved.username)

    def test_retrieve_user_email(self):
        user_retrieved = self.client.user(username='testuser')
        email_retrieved = user_retrieved.email
        self.assertEqual('a@b.nl', email_retrieved)

    def test_retrieve_user_language(self):
        user_retrieved = self.client.user(username='testuser')
        language_retrieved = user_retrieved.language
        self.assertIn(language_retrieved, LanguageCodes.values())



from pykechain.enums import (
    PropertyRepresentation,
    SelectListRepresentations,
    LinkTargets,
    FontAwesomeMode,
)
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.input_checks import check_type
from pykechain.models.representations.representation_base import BaseRepresentation


class DecimalPlaces(BaseRepresentation):
    """Representation for floating-point value properties."""

    rtype = PropertyRepresentation.DECIMAL_PLACES
    _config_value_key = "amount"

    def validate_representation(self, value: int) -> None:
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int
        :return: None
        """
        if not isinstance(value, int):
            raise IllegalArgumentError(
                '{} value "{}" is not correct: not an integer'.format(
                    self.__class__.__name__, value
                )
            )


class SignificantDigits(DecimalPlaces):
    """Representation for floating-point value properties."""

    rtype = PropertyRepresentation.SIGNIFICANT_DIGITS


class ThousandsSeparator(BaseRepresentation):
    """Representation for integer or floating-point value properties."""

    rtype = PropertyRepresentation.THOUSANDS_SEPARATOR

    def validate_representation(self, value):
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int or float
        :return: None
        """
        if not (isinstance(value, type(None))):
            raise IllegalArgumentError(
                '{} value "{}" is not correct: not NoneType'.format(
                    self.__class__.__name__, value
                )
            )


class LinkTarget(BaseRepresentation):
    """Representation for HTML link reference properties."""

    rtype = PropertyRepresentation.LINK_TARGET
    _config_value_key = "target"

    def validate_representation(self, value: LinkTargets) -> None:
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: LinkTargets
        :return: None
        """
        if value not in LinkTargets.values():
            raise IllegalArgumentError(
                '{} value "{}" is not correct: Not a CardWidgetLinkTarget option.'.format(
                    self.__class__.__name__, value
                )
            )


class ButtonRepresentation(BaseRepresentation):
    """Representation for single-select list properties."""

    rtype = PropertyRepresentation.BUTTON
    _config_value_key = PropertyRepresentation.BUTTON

    def validate_representation(self, value: SelectListRepresentations) -> None:
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: SelectListRepresentations
        :return: None
        """
        if value not in SelectListRepresentations.values():
            raise IllegalArgumentError(
                '{} value "{}" is not correct: Not a SelectListRepresentations option.'.format(
                    self.__class__.__name__, value
                )
            )


class Autofill(BaseRepresentation):
    """Representation for date(time) properties."""

    rtype = PropertyRepresentation.AUTOFILL
    _config_value_key = PropertyRepresentation.AUTOFILL

    def validate_representation(self, value: bool) -> None:
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: bool
        :return: None
        """
        check_type(value, bool, "autofill")


class CustomIconRepresentation(BaseRepresentation):
    """Representation for scope and activities to display a custom Font Awesome icon."""

    rtype = "customIcon"
    _config_value_key = "displayIcon"
    _display_mode_key = "displayIconMode"

    def __init__(self, *args, **kwargs):
        """
        Create a custom icon representation.

        Display mode of the icon will be `regular` by default.
        """
        super().__init__(*args, **kwargs)
        if self._display_mode_key not in self._config:
            self._config[self._display_mode_key] = FontAwesomeMode.REGULAR

    def validate_representation(self, value: str):
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: str
        :return: None
        """
        if not isinstance(value, str):
            raise IllegalArgumentError(
                '{} value "{}" is not correct: not a string'.format(
                    self.__class__.__name__, value
                )
            )

    @property
    def display_mode(self):
        """Get the the display mode of the custom icon representation."""
        return self._config[self._display_mode_key]

    @display_mode.setter
    def display_mode(self, mode: FontAwesomeMode) -> None:
        """
        Set the the display mode of the custom icon representation.

        :param mode: FontAwesome display mode
        :type mode: FontAwesomeMode
        """
        if mode not in set(FontAwesomeMode.values()):
            raise IllegalArgumentError(
                '{} mode "{}" is not a FontAwesomeMode option.'.format(
                    self.__class__.__name__, mode
                )
            )

        self._config[self._display_mode_key] = mode
        self.value = self.value  # trigger update



import io
import json
import os

import requests
from typing import Text, Any, Optional

from pykechain.exceptions import APIError
from pykechain.models.property2_attachment import AttachmentProperty2
from pykechain.models.property import Property


class AttachmentProperty(Property, AttachmentProperty2):
    """A virtual object representing a KE-chain attachment property."""

    @property
    def value(self):
        """Retrieve the data value of this attachment.

        Will show the filename of the attachment if there is an attachment available otherwise None
        Use save_as in order to download as a file.

        Example
        -------
        >>> file_attachment_property = project.part('Bike').property('file_attachment')
        >>> if file_attachment_property.value:
        ...     file_attachment_property.save_as('file.ext')
        ... else:
        ...     print('file attachment not set, its value is None')

        """
        if self.has_value():
            return "[Attachment: {}]".format(self.filename)
        else:
            return None

    @value.setter
    def value(self, value):
        if value is None:
            self.clear()
        else:
            self.upload(data=value)

    def clear(self) -> None:
        """Clear the attachment from the attachment field.

        :raises APIError: if unable to remove the attachment
        """
        if self._put_value(None) is None:
            self._value = None
            self._json_data['value'] = None

    @property
    def filename(self) -> Optional[Text]:
        """Filename of the attachment, without the full 'attachment' path."""
        return self._value.split('/')[-1] if self.has_value() else None

    def json_load(self):
        """Download the data from the attachment and deserialise the contained json.

        :return: deserialised json data as :class:`dict`
        :raises APIError: When unable to retrieve the json from KE-chain
        :raises JSONDecodeError: When there was a problem in deserialising the json

        Example
        -------
        Ensure that the attachment is valid json data

        >>> json_attachment = project.part('Bike').property('json_attachment')
        >>> deserialised_json = json_attachment.json_load()

        """
        return self._download().json()

    def upload(self, data: Any, **kwargs: Any) -> None:
        """Upload a file to the attachment property.

        When providing a :class:`matplotlib.figure.Figure` object as data, the figure is uploaded as PNG.
        For this, `matplotlib`_ should be installed.

        :param data: File path
        :type data: basestring
        :raises APIError: When unable to upload the file to KE-chain
        :raises OSError: When the path to the file is incorrect or file could not be found

        .. _matplotlib: https://matplotlib.org/
        """
        try:
            import matplotlib.figure

            if isinstance(data, matplotlib.figure.Figure):
                self._upload_plot(data, **kwargs)
                return
        except ImportError:
            pass

        if isinstance(data, str):
            with open(data, 'rb') as fp:
                self._upload(fp)
        else:
            self._upload_json(data, **kwargs)
        self._value = data

    def save_as(self, filename: Optional[Text] = None) -> None:
        """Download the attachment to a file.

        :param filename: (optional) File path. If not provided, will be saved to current working dir
                         with `self.filename`.
        :type filename: basestring or None
        :raises APIError: When unable to download the data
        :raises OSError: When unable to save the data to disk
        """
        filename = filename or os.path.join(os.getcwd(), self.filename)

        with open(filename, 'w+b') as f:
            for chunk in self._download():
                f.write(chunk)

    def _upload_json(self, content, name='data.json'):
        data = (name, json.dumps(content), 'application/json')

        self._upload(data)

    def _upload_plot(self, figure, name='plot.png'):
        buffer = io.BytesIO()

        figure.savefig(buffer, format="png")

        data = (name, buffer.getvalue(), 'image/png')

        self._upload(data)
        self._value = name

    # custom for PIM2
    def _download(self):
        url = self._client._build_url('property_download', property_id=self.id)

        response = self._client._request('GET', url)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not download property value.", response=response)

        return response

    # custom for PIM2
    def _upload(self, data):
        url = self._client._build_url('property_upload', property_id=self.id)

        response = self._client._request('POST', url,
                                         data={"part": self._json_data['part_id']},
                                         files={"attachment": data})

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not upload attachment", response=response)



import datetime

import pytz

from pykechain.exceptions import APIError, NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models.banner import Banner
from tests.classes import TestBetamax

NEW_YEAR_2020_naive = datetime.datetime.strptime('2020-01-01T00:01:00', '%Y-%m-%dT%H:%M:%S')
NEW_YEAR_2020 = NEW_YEAR_2020_naive.replace(tzinfo=pytz.UTC)


class TestBanners(TestBetamax):
    TEXT = '__This is a test banner__'
    URL = 'https://www.ke-chain.nl/'
    ICON = 'poo-storm'

    KWARGS = dict(
        text=TEXT,
        icon=ICON,
        active_from=NEW_YEAR_2020,
    )

    def setUp(self):
        super().setUp()
        self.banner = self.client.create_banner(
            url=self.URL,
            is_active=True,
            **self.KWARGS
        )  # type: Banner

    def tearDown(self):
        if self.banner:
            try:
                self.banner.delete()
            except APIError:
                pass
        super().tearDown()

    def test_create(self):
        self.assertIsInstance(self.banner, Banner)
        self.assertEqual(self.TEXT, self.banner.text)
        self.assertEqual(self.URL, self.banner.url)
        self.assertEqual(self.ICON, self.banner.icon)
        self.assertTrue(self.banner.is_active)

    def test_create_empty(self):
        self.banner.delete()
        self.banner = self.client.create_banner(**self.KWARGS)

    def test_create_invalid_inputs(self):
        self.banner.delete()
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(text=3, icon=self.ICON, active_from=NEW_YEAR_2020)
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(text='some text', icon=True, active_from=NEW_YEAR_2020)
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(text='some text', icon=self.ICON, active_from='2020-01-01T00:01:00')
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(active_until='later', **self.KWARGS)
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(is_active=1, **self.KWARGS)
        with self.assertRaises(IllegalArgumentError):
            self.banner = self.client.create_banner(url='www.ke-chain/scopes/', **self.KWARGS)

    def test_delete(self):
        banner_id = self.banner.id
        self.banner.delete()

        with self.assertRaises(APIError, msg='Cant delete the same banner twice.'):
            self.banner.delete()
        with self.assertRaises(NotFoundError, msg='Banner is still found!'):
            self.client.banner(pk=banner_id)

    def test_get_banners(self):
        all_banners = self.client.banners()

        self.assertIsInstance(all_banners, list)
        self.assertTrue(all_banners)

        first_banner = all_banners[0]

        self.assertIsInstance(first_banner, Banner)

    def test_get_banners_invalid_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.client.banners(pk=1)
        with self.assertRaises(IllegalArgumentError):
            self.client.banners(text=False)
        with self.assertRaises(IllegalArgumentError):
            self.client.banners(is_active="")

    def test_get_banner(self):
        banner = self.client.banner(text=self.TEXT)

        self.assertTrue(banner)
        self.assertIsInstance(banner, Banner)

        with self.assertRaises(MultipleFoundError):
            self.client.banner()

    def test_get_active_banner(self):
        active_banner = self.client.active_banner()

        self.assertIsNotNone(active_banner)
        self.assertIsInstance(active_banner, Banner)
        self.assertEqual(self.banner, active_banner)

    def test_edit(self):
        text = '__RENAMED BANNER'
        icon = 'site-map'
        url = 'https://www.google.com/'
        later = NEW_YEAR_2020 + datetime.timedelta(hours=1)

        self.banner.edit(
            text=text,
            icon=icon,
            active_from=NEW_YEAR_2020,
            active_until=later,
            is_active=False,
            url=url,
        )

        self.assertEqual(text, self.banner.text)
        self.assertEqual(icon, self.banner.icon)
        self.assertEqual(NEW_YEAR_2020, self.banner.active_from)
        self.assertEqual(later, self.banner.active_until)
        self.assertFalse(self.banner.is_active)
        self.assertEqual(url, self.banner.url)

    def test_edit_single_inputs(self):
        for kwarg in [
            dict(text='New text'),
            dict(icon='gifts'),
            dict(active_from=NEW_YEAR_2020 - datetime.timedelta(days=1)),
            dict(active_until=NEW_YEAR_2020 + datetime.timedelta(weeks=1)),
            dict(is_active=False),
            dict(url='https://www.google.com'),
        ]:
            with self.subTest(msg=kwarg):
                self.banner.edit(**kwarg)

    # noinspection PyTypeChecker
    def test_edit_invalid_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(text=True)
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(icon=50)
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(active_from='now')
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(active_until='later')
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(url='ke-chain,com')
        with self.assertRaises(IllegalArgumentError):
            self.banner.edit(is_active='Yes')

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_banner_clear_values(self):
        # setup
        initial_text = 'Happy new Year'
        initial_icon = 'gifts'
        initial_active_from = NEW_YEAR_2020 - datetime.timedelta(days=1)
        initial_active_until = NEW_YEAR_2020 + datetime.timedelta(days=1)
        initial_is_active = False
        initial_url = 'https://www.google.com'

        self.banner.edit(
            text=initial_text,
            icon=initial_icon,
            active_from=NEW_YEAR_2020 - datetime.timedelta(days=1),
            active_until=NEW_YEAR_2020 + datetime.timedelta(days=1),
            is_active=initial_is_active,
            url=initial_url
        )

        # Edit without mentioning values, everything should stay the same
        new_text = '2021!!!'
        new_is_active = True

        self.banner.edit(
            text=new_text,
            is_active=new_is_active,
        )

        # testing
        self.assertEqual(self.banner.text, new_text)
        self.assertEqual(self.banner.icon, initial_icon)
        self.assertEqual(self.banner.active_from, initial_active_from)
        self.assertEqual(self.banner.active_until, initial_active_until)
        self.assertEqual(self.banner.is_active, new_is_active)
        self.assertEqual(self.banner.url, initial_url)

        # Edit with clearing the values, name and status cannot be cleared
        self.banner.edit(
            text=None,
            icon=None,
            active_from=None,
            active_until=None,
            is_active=None,
            url=None
        )

        self.assertEqual(self.banner.text, new_text)
        self.assertEqual(self.banner.icon, str())
        self.assertEqual(self.banner.active_from, initial_active_from)
        self.assertEqual(self.banner.active_until, initial_active_until)
        self.assertEqual(self.banner.is_active, new_is_active)
        self.assertEqual(self.banner.url, str())



from pykechain.enums import ValidatorEffectTypes
from pykechain.models.validators.validators_base import ValidatorEffect


class TextEffect(ValidatorEffect):
    """A Text effect, that will set a text.

    .. versionadded:: 2.2
    """

    effect = ValidatorEffectTypes.TEXT_EFFECT

    def __init__(self, json=None, text="The validation resulted in an error.", **kwargs):
        """Construct an helptext effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param text: (optional) text to provide, default empty
        :type text: basestring
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super(TextEffect, self).__init__(json=json, text=text, **kwargs)
        self.text = text

    def as_json(self):
        # type: () -> dict
        """Represent effect as JSON dict."""
        self._config['text'] = self.text
        return self._json


class ErrorTextEffect(ValidatorEffect):
    """A Errortext effect, that will set a text.

    .. versionadded:: 2.2
    """

    effect = ValidatorEffectTypes.ERRORTEXT_EFFECT

    def __init__(self, json=None, text="The validation resulted in an error.", **kwargs):
        """Construct an errortext effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param text: (optional) text to provide, default empty
        :type text: basestring
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super(ErrorTextEffect, self).__init__(json=json, **kwargs)


class HelpTextEffect(ValidatorEffect):
    """A Errortext effect, that will set a text.

    .. versionadded:: 2.2
    """

    effect = ValidatorEffectTypes.HELPTEXT_EFFECT

    def __init__(self, json=None, text="", **kwargs):
        """Construct an helptext effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param text: (optional) text to provide, default empty
        :type text: basestring
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super(HelpTextEffect, self).__init__(json=json, text=text, **kwargs)


class VisualEffect(ValidatorEffect):
    """A visualeffect, to be processed by the frontend.

    .. versionadded:: 2.2

    :ivar applyCss: css class to apply in case of this effect
    """

    effect = ValidatorEffectTypes.VISUALEFFECT

    def __init__(self, json=None, applyCss=None, **kwargs):
        """Construct an visual effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param applyCss: (optional) class to apply, defaults to 'valid'
        :type applyCss: basestring
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super(VisualEffect, self).__init__(json=json, **kwargs)
        self.applyCss = applyCss or self._config.get('applyCss')

    def as_json(self):
        # type: () -> dict
        """Represent effect as JSON dict."""
        self._config['applyCss'] = self.applyCss
        self._json['config'] = self._config
        return self._json


class ValidVisualEffect(VisualEffect):
    """Effect that may apply a css class when the result of the validator is valid.

    .. versionadded:: 2.2
    """

    def __init__(self, json=None, applyCss='valid', **kwargs):
        """Construct an valid visual effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param applyCss: (optional) class to apply, defaults to 'valid'
        :type applyCss: basestring
        """
        super(ValidVisualEffect, self).__init__(json=json, applyCss=applyCss, **kwargs)


class InvalidVisualEffect(VisualEffect):
    """Effect that may apply a css class when the result of the validator is invalid.

    .. versionadded:: 2.2
    """

    def __init__(self, json=None, applyCss='invalid', **kwargs):
        """Construct an invalid visual effect.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param applyCss: (optional) class to apply, defaults to 'invalid'
        :type applyCss: basestring
        """
        super(InvalidVisualEffect, self).__init__(json=json, applyCss=applyCss, **kwargs)



import warnings
from typing import List, Optional, Text, Union, Any, Tuple

from pykechain.defaults import PARTS_BATCH_LIMIT
from pykechain.enums import Category, FilterType
from pykechain.models.input_checks import check_type
from pykechain.models.property2_multi_reference import MultiReferenceProperty2
from pykechain.models.base_reference import _ReferencePropertyInScope
from pykechain.models.part import Part
from pykechain.models.value_filter import PropertyValueFilter
from pykechain.models.widgets.enums import MetaWidget
from pykechain.models.widgets.helpers import _check_prefilters, _check_excluded_propmodels
from pykechain.utils import get_in_chunks


class MultiReferenceProperty(_ReferencePropertyInScope, MultiReferenceProperty2):
    """A virtual object representing a KE-chain multi-references property.

    .. versionadded:: 1.14
    """

    REFERENCED_CLASS = Part

    def _retrieve_objects(self, **kwargs) -> List[Part]:
        """
        Retrieve a list of Parts.

        :param kwargs: optional inputs
        :return: list of Part objects
        """
        part_ids = self._validate_values()

        parts = []

        if part_ids:
            if self.category == Category.MODEL:
                parts = [self._client.part(pk=part_ids[0], category=None)]
            elif self.category == Category.INSTANCE:
                # Retrieve the referenced model for low-permissions scripts to enable use of the `id__in` key
                if False:  # TODO Check for script permissions in order to skip the model() retrieval
                    models = [None]
                else:
                    models = self.model().value
                if models:
                    parts = list()
                    for chunk in get_in_chunks(part_ids, PARTS_BATCH_LIMIT):
                        parts.extend(
                            list(self._client.parts(
                                id__in=','.join(chunk),
                                model=models[0],
                                category=None,
                            ))
                        )
        return parts

    def choices(self) -> List[Part]:
        """Retrieve the parts that you can reference for this `MultiReferenceProperty`.

        This method makes 2 API calls: 1) to retrieve the referenced model, and 2) to retrieve the instances of
        that model.

        :return: the :class:`Part`'s that can be referenced as a :class:`~pykechain.model.PartSet`.
        :raises APIError: When unable to load and provide the choices

        Example
        -------
        >>> reference_property = project.part('Bike').property('a_multi_reference_property')
        >>> referenced_part_choices = reference_property.choices()

        """
        possible_choices = list()
        # Check whether the model of this reference property (possible itself) has a configured value
        if self.model().has_value():
            # If a model is configured, retrieve its ID
            choices_model_id = self.model()._value[0].get('id')

            # Determine which parts are filtered out
            prefilter = self._options.get(MetaWidget.PREFILTERS, {}).\
                get(MetaWidget.PROPERTY_VALUE_PREFILTER)  # type: Optional[Text]

            # Retrieve all part instances with this model ID
            possible_choices = self._client.parts(
                model_id=choices_model_id,
                property_value=prefilter,
            )

        return possible_choices

    def set_prefilters(
            self,
            property_models: List[Union[Text, 'AnyProperty']] = None,
            values: List[Any] = None,
            filters_type: List[FilterType] = None,
            prefilters: List[PropertyValueFilter] = None,
            overwrite: Optional[bool] = False,
            clear: Optional[bool] = False,
            validate: Optional[Union[bool, Part]] = True,
    ) -> None:
        """
        Set the pre-filters on a `MultiReferenceProperty`.

        :param property_models: `list` of `Property` models (or their IDs) to set pre-filters on
        :type property_models: list
        :param values: `list` of values to pre-filter on, value has to match the property type.
        :type values: list
        :param filters_type: `list` of filter types per pre-filter, one of :class:`enums.FilterType`,
                defaults to `FilterType.CONTAINS`
        :type filters_type: list
        :param prefilters: `list` of PropertyValueFilter objects
        :type prefilters: list
        :param overwrite: whether existing pre-filters should be overwritten, if new filters to the same property
            are provided as input. Does not remove non-conflicting prefilters. (default = False)
        :type overwrite: bool
        :param clear: whether all existing pre-filters should be cleared. (default = False)
        :type clear: bool
        :param validate: whether the pre-filters are validated to the referenced model, which can be provided as well
        :type validate: bool

        :raises IllegalArgumentError: when the type of the input is provided incorrect.
        """
        if not clear:
            list_of_prefilters = PropertyValueFilter.parse_options(options=self._options)
        else:
            list_of_prefilters = list()

        if prefilters is None:
            new_prefilters = {
                "property_models": property_models,
                "values": values,
                "filters_type": filters_type,
            }
        else:
            new_prefilters = prefilters

        if not validate:
            part_model = None
        elif isinstance(validate, Part):
            part_model = validate
        else:
            part_model = self.value[0]

        verified_prefilters = _check_prefilters(
            part_model=part_model,
            prefilters=new_prefilters,
        )

        if overwrite:  # Remove pre-filters from the existing prefilters if they match the property model UUID
            provided_filter_ids = {pf.id for pf in verified_prefilters}
            list_of_prefilters = [pf for pf in list_of_prefilters if pf.id not in provided_filter_ids]

        list_of_prefilters += verified_prefilters

        # Only update the options if there are any prefilters to be set, or if the original filters have to overwritten
        if list_of_prefilters or clear:
            options_to_set = self._options

            # Always create the entire prefilter json structure
            options_to_set["prefilters"] = {
                "property_value": ",".join([pf.format() for pf in list_of_prefilters])
            }

            self.edit(options=options_to_set)

    def get_prefilters(
            self,
            as_lists: Optional[bool] = False,
    ) -> Union[
        List[PropertyValueFilter],
        Tuple[List[Text]]
    ]:
        """
        Retrieve the pre-filters applied to the reference property.

        :param as_lists: (O) (default = False)
            If True, the pre-filters are returned as three lists of property model UUIDs, values and filter types.
        :return: prefilters
        """
        check_type(as_lists, bool, "as_lists")

        prefilters = PropertyValueFilter.parse_options(options=self._options)

        if as_lists:
            property_model_ids = [pf.id for pf in prefilters]
            values = [pf.value for pf in prefilters]
            filter_types = [pf.type for pf in prefilters]
            prefilters = tuple([property_model_ids, values, filter_types])

            warnings.warn(
                "Prefilters will be provided as list of `PropertyValueFilter` objects. "
                "Separate lists will be deprecated in January 2021.",  # TODO Deprecate January 2021
                PendingDeprecationWarning,
            )

        return prefilters

    def set_excluded_propmodels(
            self,
            property_models: List[Union[Text, 'AnyProperty']],
            overwrite: Optional[bool] = False,
            validate: Optional[Union[bool, Part]] = True,
    ) -> None:
        """
        Exclude a list of properties from being visible in the part-shop and modal (pop-up) of the reference property.

        :param property_models: `list` of Property models (or their IDs) to exclude.
        :type property_models: list
        :param overwrite: flag whether to overwrite existing (True) or append (False, default) to existing filter(s).
        :type overwrite bool
        :param validate: whether the pre-filters are validated to the referenced model, which can be provided as well
        :type validate: bool

        :raises IllegalArgumentError
        """
        if not overwrite:
            list_of_propmodels_excl = self._options.get('propmodels_excl', [])
        else:
            list_of_propmodels_excl = list()

        if not validate:
            part_model = None
        elif isinstance(validate, Part):
            part_model = validate
        else:
            part_model = self.value[0]

        list_of_propmodels_excl.extend(_check_excluded_propmodels(
            part_model=part_model,
            property_models=property_models,
        ))

        options_to_set = self._options
        options_to_set['propmodels_excl'] = list(set(list_of_propmodels_excl))

        self.edit(options=options_to_set)

    def get_excluded_propmodel_ids(self) -> List[Text]:
        """
        Retrieve a list of property model UUIDs which are not visible.

        :return: list of UUIDs
        :rtype list
        """
        return self._options.get("propmodels_excl", [])



from pykechain.exceptions import _DeprecationMixin


class Activity2(_DeprecationMixin):
    """A virtual object representing a KE-chain activity.

    .. versionadded:: 2.0

    :ivar id: id of the activity
    :type id: uuid
    :ivar name: name of the activity
    :type name: basestring
    :ivar created_at: created datetime of the activity
    :type created_at: datetime
    :ivar updated_at: updated datetime of the activity
    :type updated_at: datetime
    :ivar description: description of the activity
    :type description: basestring
    :ivar status: status of the activity. One of :class:`pykechain.enums.ActivityStatus`
    :type status: basestring
    :ivar classification: classification of the activity. One of :class:`pykechain.enums.ActivityClassification`
    :type classification: basestring
    :ivar activity_type: Type of the activity. One of :class:`pykechain.enums.ActivityType` for WIM version 2
    :type activity_type: basestring
    """

    pass



from datetime import datetime
from typing import Union, Text, Dict, Optional, List  # noqa: F401

import requests

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import (
    Multiplicity,
    ScopeStatus,
    SubprocessDisplayMode,
    KEChainPages,
    ScopeRoles,
    ScopeMemberActions,
    ScopeCategory,
)
from pykechain.exceptions import APIError, NotFoundError, IllegalArgumentError
from pykechain.models.scope2 import Scope2
from pykechain.models.service import Service, ServiceExecution
from pykechain.models.activity import Activity
from pykechain.models.part import Part
from pykechain.models.base import Base
from pykechain.models.property import Property
from pykechain.models.input_checks import (
    check_text,
    check_datetime,
    check_enum,
    check_list_of_text,
    check_base,
    check_type,
)
from pykechain.models.representations import BaseRepresentation
from pykechain.models.representations.component import RepresentationsComponent
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from pykechain.models.tags import TagsMixin
from pykechain.models.team import Team
from pykechain.utils import parse_datetime, find, Empty, clean_empty_values, empty


class Scope(Base, TagsMixin, Scope2):
    """A virtual object representing a KE-chain scope.

    :ivar id: id of the activity
    :type id: uuid
    :ivar name: name of the activity
    :type name: basestring
    :ivar created_at: created datetime of the activity
    :type created_at: datetime
    :ivar updated_at: updated datetime of the activity
    :type updated_at: datetime
    :ivar description: description of the activity
    :type description: basestring
    :ivar workflow_root: uuid of the workflow root object
    :type workflow_root: uuid
    :ivar status: status of the scope. One of :class:`pykechain.enums.ScopeStatus`
    :type status: basestring
    :ivar type: Type of the Scope. One of :class:`pykechain.enums.ScopeType` for WIM version 2
    :type type: basestring
    """

    def __init__(self, json: Dict, **kwargs) -> None:
        """Construct a scope from provided json data."""
        super().__init__(json, **kwargs)

        # for 'kechain2.core.wim <2.0.0'
        self.process = json.get("process")
        # for 'kechain2.core.wim >=2.0.0'
        self.workflow_root = json.get("workflow_root_id")

        self._workflow_root_process = None
        self._catalog_root_process = None
        self._app_root_process = None
        self._product_root_model = None
        self._product_root_instance = None
        self._catalog_root_model = None
        self._catalog_root_instance = None

        self.ref = json.get("ref")
        self.description = json.get("text")
        self.status = json.get("status")
        self.category = json.get("category")

        self._tags = json.get("tags")

        self.start_date = parse_datetime(json.get("start_date"))
        self.due_date = parse_datetime(json.get("due_date"))

        self._representations_container = RepresentationsComponent(
            self, self.options.get("representations", {}), self._save_representations,
        )

    @property
    def team(self) -> Optional[Team]:
        """Team to which the scope is assigned."""
        team_dict = self._json_data.get("team_id_name")
        if team_dict and team_dict.get("id"):
            return self._client.team(pk=team_dict.get("id"))
        else:
            return None

    @property
    def options(self) -> Dict:
        """Options of the Scope.

        .. versionadded: 3.0
        """
        return self._json_data.get("scope_options")

    @options.setter
    def options(self, option_value):
        self.edit(options=option_value)

    def refresh(self, json=None, url=None, extra_params=None):
        """Refresh the object in place."""
        super().refresh(
            json=json,
            url=self._client._build_url("scope", scope_id=self.id),
            extra_params=API_EXTRA_PARAMS["scope"],
        )

    @property
    def representations(self) -> List['BaseRepresentation']:
        """Get and set the scope representations."""
        return self._representations_container.get_representations()

    @representations.setter
    def representations(self, value):
        self._representations_container.set_representations(value)

    def _save_representations(self, representation_options):
        options = self.options
        options.update({"representations": representation_options})
        self.options = options

    @property
    def workflow_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification WORKFLOW."""
        if self._workflow_root_process is None:
            self._workflow_root_process = self.activity(
                id=self._json_data["workflow_root_id"]
            )
        return self._workflow_root_process

    @property
    def app_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification APP."""
        if self._app_root_process is None:
            self._app_root_process = self.activity(id=self._json_data["app_root_id"])
        return self._app_root_process

    @property
    def catalog_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification CATALOG."""
        if self._catalog_root_process is None:
            self._catalog_root_process = self.activity(
                id=self._json_data["catalog_root_id"]
            )
        return self._catalog_root_process

    @property
    def product_root_model(self) -> "Part":
        """Retrieve the Part root object with classification PRODUCT and category MODEL."""
        if self._product_root_model is None:
            self._product_root_model = self.model(
                id=self._json_data["product_model_id"]
            )
        return self._product_root_model

    @property
    def product_root_instance(self) -> "Part":
        """Retrieve the Part root object with classification PRODUCT and category INSTANCE."""
        if self._product_root_instance is None:
            self._product_root_instance = self.part(
                id=self._json_data["product_instance_id"]
            )
        return self._product_root_instance

    @property
    def catalog_root_model(self) -> "Part":
        """Retrieve the Part root object with classification CATALOG and category MODEL."""
        if self._catalog_root_model is None:
            self._catalog_root_model = self.model(
                id=self._json_data["catalog_model_id"]
            )
        return self._catalog_root_model

    @property
    def catalog_root_instance(self) -> "Part":
        """Retrieve the Part root object with classification CATALOG and category INSTANCE."""
        if self._catalog_root_instance is None:
            self._catalog_root_instance = self.part(
                id=self._json_data["catalog_instance_id"]
            )
        return self._catalog_root_instance

    #
    # CRUD methods
    #

    def _update_scope_project_team(self, action, role, user):
        """
        Update the Project Team of the Scope. Updates include addition or removing of managers or members.

        :param action: type of action to be applied
        :type action: ScopeMemberActions
        :param role: type of role to be applied to the user
        :type role: ScopeRoles
        :param user: the username of the user to which the action applies to
        :type user: basestring
        :raises APIError: When unable to update the scope project team.
        """
        action = check_enum(action, ScopeMemberActions, "action")
        role = check_enum(role, ScopeRoles, "role")
        user = check_text(user, "user")

        users = self._client._retrieve_users()["results"]  # type: List[Dict]
        user_object = find(users, lambda u: u["username"] == user)  # type: Dict
        if user_object is None:
            raise NotFoundError('User "{}" does not exist'.format(user))

        url = self._client._build_url(
            "scope_{}_{}".format(action, role), scope_id=self.id
        )

        response = self._client._request(
            "PUT",
            url,
            params=API_EXTRA_PARAMS[self.__class__.__name__.lower()],
            data={"user_id": user_object["pk"]},
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(
                "Could not {} {} in Scope".format(action, role), response=response
            )

        self.refresh(json=response.json().get("results")[0])

    def edit(
            self,
            name: Optional[Union[Text, Empty]] = empty,
            description: Optional[Union[Text, Empty]] = empty,
            start_date: Optional[Union[datetime, Empty]] = empty,
            due_date: Optional[Union[datetime, Empty]] = empty,
            status: Optional[Union[Text, ScopeStatus, Empty]] = empty,
            category: Optional[Union[Text, ScopeCategory, Empty]] = empty,
            tags: Optional[Union[List[Text], Empty]] = empty,
            team: Optional[Union[Team, Text, Empty]] = empty,
            options: Optional[Union[Dict, Empty]] = empty,
            **kwargs
    ) -> None:
        """
        Edit the details of a scope.

        Setting an input to None will clear out the value (exception being name and status).

        :param name: (optionally) edit the name of the scope. Name cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optionally) edit the description of the scope or clear it
        :type description: basestring or None or Empty
        :param start_date: (optionally) edit the start date of the scope as a datetime object (UTC time/timezone
                            aware preferred) or clear it
        :type start_date: datetime or None or Empty
        :param due_date: (optionally) edit the due_date of the scope as a datetime object (UTC time/timzeone
                            aware preferred) or clear it
        :type due_date: datetime or None or Empty
        :param status: (optionally) edit the status of the scope as a string based. Status cannot be cleared.
        :type status: ScopeStatus or Empty
        :param category (optionally) edit the category of the scope
        :type category: ScopeCategory or Empty
        :param tags: (optionally) replace the tags on a scope, which is a list of strings ["one","two","three"] or
                    clear them
        :type tags: list of basestring or None or Empty
        :param team: (optionally) add the scope to a team. Team cannot be cleared.
        :type team: UUIDstring or None or Empty
        :param options: (optionally) custom options dictionary stored on the scope object
        :type options: dict or None or Empty

        :raises IllegalArgumentError: if the type of the inputs is not correct
        :raises APIError: if another Error occurs
        :warns: UserWarning - When a naive datetime is provided. Defaults to UTC.

        Examples
        --------
        >>> from datetime import datetime
        >>> project.edit(name='New project name',
        ...              description='Changing the description just because I can',
        ...              start_date=datetime.now(),  # naive time is interpreted as UTC time
        ...              status=ScopeStatus.CLOSED)

        If we want to provide timezone aware datetime objects we can use the 3rd party convenience library :mod:`pytz`.
        Mind that we need to fetch the timezone first and use `<timezone>.localize(<your datetime>)` to make it
        work correctly.

        Using `datetime(2017,6,1,23,59,0 tzinfo=<tz>)` does NOT work for most timezones with a
        daylight saving time. Check the `pytz <http://pythonhosted.org/pytz/#localized-times-and-date-arithmetic>`_
        documentation.

        To make it work using :mod:`pytz` and timezone aware :mod:`datetime` see the following example::

        >>> import pytz
        >>> start_date_tzaware = datetime.now(pytz.utc)
        >>> mytimezone = pytz.timezone('Europe/Amsterdam')
        >>> due_date_tzaware = mytimezone.localize(datetime(2019, 10, 27, 23, 59, 0))
        >>> project.edit(due_date=due_date_tzaware, start_date=start_date_tzaware)

        To assign a scope to a team see the following example::

        >>> my_team = client.team(name='My own team')
        >>> project.edit(team=my_team)

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the due_date, but leave everything else
        unchanged.

        >>> project.edit(due_date=None)

        """
        update_dict = {
            "id": self.id,
            "name": check_text(name, "name") or self.name,
            "text": check_text(description, "description") or str(),
            "start_date": check_datetime(start_date, "start_date"),
            "due_date": check_datetime(due_date, "due_date"),
            "status": check_enum(status, ScopeStatus, "status") or self.status,
            "category": check_enum(category, ScopeCategory, "category"),
            "tags": check_list_of_text(tags, "tags", True) or list(),
            "team_id": check_base(team, Team, "team") or str(),
            "scope_options": check_type(options, dict, "options") or dict(),
        }

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url("scope", scope_id=self.id)

        response = self._client._request(
            "PUT",
            url,
            params=API_EXTRA_PARAMS[self.__class__.__name__.lower()],
            json=update_dict,
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Scope {}".format(self), response=response)

        self.refresh(json=response.json().get("results")[0])

        # TODO tags that are set are not in response
        if tags is not None and not isinstance(tags, Empty):
            self._tags = tags

    def clone(self, **kwargs) -> 'Scope':
        """Clone a scope.

        See :method:`pykechain.Client.clone_scope()` for available parameters.
        """
        return self._client.clone_scope(source_scope=self, **kwargs)

    def delete(self, asynchronous=True):
        """Delete the scope.

        Only works with enough permissions.

        .. versionadded: 3.0

        See :method:`pykechain.Client.delete_scope()` for available parameters.
        :raises ForbiddenError: if you do not have the permissions to delete a scope
        """
        return self._client.delete_scope(scope=self, asynchronous=asynchronous)

    #
    # Part methods
    #

    def parts(self, *args, **kwargs) -> List['Part']:
        """Retrieve parts belonging to this scope.

        This uses

        See :class:`pykechain.Client.parts` for available parameters.
        """
        return self._client.parts(*args, scope_id=self.id, **kwargs)

    def part(self, *args, **kwargs) -> 'Part':
        """Retrieve a single part belonging to this scope.

        See :class:`pykechain.Client.part` for available parameters.
        """
        return self._client.part(*args, scope_id=self.id, **kwargs)

    def properties(self, *args, **kwargs) -> List['Property']:
        """Retrieve properties belonging to this scope.

        .. versionadded: 3.0

        See :class:`pykechain.Client.properties` for available parameters.
        """
        return self._client.properties(*args, scope_id=self.id, **kwargs)

    def property(self, *args, **kwargs) -> 'Property':
        """Retrieve a single property belonging to this scope.

        .. versionadded: 3.0

        See :class:`pykechain.Client.property` for available parameters.
        """
        return self._client.property(*args, scope_id=self.id, **kwargs)

    def model(self, *args, **kwargs) -> 'Part':
        """Retrieve a single model belonging to this scope.

        See :class:`pykechain.Client.model` for available parameters.
        """
        return self._client.model(*args, scope_id=self.id, **kwargs)

    def create_model(self, parent, name, multiplicity=Multiplicity.ZERO_MANY) -> 'Part':
        """Create a single part model in this scope.

        See :class:`pykechain.Client.create_model` for available parameters.
        """
        return self._client.create_model(parent, name, multiplicity=multiplicity)

    def create_model_with_properties(
        self,
        parent,
        name,
        multiplicity=Multiplicity.ZERO_MANY,
        properties_fvalues=None,
        **kwargs
    ) -> 'Part':
        """Create a model with its properties in a single API request.

        See :func:`pykechain.Client.create_model_with_properties()` for available parameters.
        """
        return self._client.create_model_with_properties(
            parent,
            name,
            multiplicity=multiplicity,
            properties_fvalues=properties_fvalues,
            **kwargs
        )

    #
    # Activity methods
    #

    def activities(self, *args, **kwargs) -> List['Activity']:
        """Retrieve activities belonging to this scope.

        See :class:`pykechain.Client.activities` for available parameters.
        """
        return self._client.activities(*args, scope=self.id, **kwargs)

    def activity(self, *args, **kwargs) -> 'Activity':
        """Retrieve a single activity belonging to this scope.

        See :class:`pykechain.Client.activity` for available parameters.
        """
        return self._client.activity(*args, scope=self.id, **kwargs)

    def create_activity(self, *args, **kwargs) -> 'Activity':
        """Create a new activity belonging to this scope.

        See :class:`pykechain.Client.create_activity` for available parameters.
        """
        return self._client.create_activity(self.workflow_root, *args, **kwargs)

    def side_bar(self, *args, **kwargs) -> Optional[SideBarManager]:
        """Retrieve the side-bar manager."""
        return SideBarManager(scope=self, *args, **kwargs)

    def set_landing_page(
        self,
        activity: Union["Activity", KEChainPages],
        task_display_mode: Optional[
            SubprocessDisplayMode
        ] = SubprocessDisplayMode.ACTIVITIES,
    ) -> None:
        """
        Update the landing page of the scope.

        :param activity: Activity object or KEChainPages option
        :type activity: (Activity, KEChainPages)
        :param task_display_mode: display mode of the activity in KE-chain
        :type task_display_mode: SubprocessDisplayMode
        :return: None
        :rtype None
        """
        from pykechain.models import Activity

        if not (isinstance(activity, Activity) or activity in KEChainPages.values()):
            raise IllegalArgumentError(
                'activity must be of class Activity or a KEChainPages option, "{}" is not.'.format(
                    activity
                )
            )

        check_enum(task_display_mode, SubprocessDisplayMode, "task_display_mode")

        if isinstance(activity, Activity):
            url = "#/scopes/{}/{}/{}".format(self.id, task_display_mode, activity.id)
        else:
            url = "#/scopes/{}/{}".format(self.id, activity)

        options = dict(self.options)
        options.update({"landingPage": url})
        self.options = options

    def get_landing_page_url(self) -> Optional[Text]:
        """
        Retrieve the landing page URL, if it is set in the options.

        :return: Landing page url
        """
        return self.options.get("landingPage")

    #
    # Service Methods
    #

    def services(self, *args, **kwargs) -> List['Service']:
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.services` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.services(*args, scope=self.id, **kwargs)

    def create_service(self, *args, **kwargs) -> 'Service':
        """Create a service to current scope.

        See :class:`pykechain.Client.create_service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.create_service(*args, scope=self.id, **kwargs)

    def service(self, *args, **kwargs) -> 'Service':
        """Retrieve a single service belonging to this scope.

        See :class:`pykechain.Client.service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service(*args, scope=self.id, **kwargs)

    def service_executions(self, *args, **kwargs) -> List['ServiceExecution']:
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.service_executions` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_executions(*args, scope=self.id, **kwargs)

    def service_execution(self, *args, **kwargs) -> 'ServiceExecution':
        """Retrieve a single service execution belonging to this scope.

        See :class:`pykechain.Client.service_execution` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_execution(*args, scope=self.id, **kwargs)

    #
    # User and Members of the Scope
    #

    def members(
        self,
        is_manager: Optional[bool] = None,
        is_supervisor: Optional[bool] = None,
        is_leadmember: Optional[bool] = None,
    ) -> List[Dict]:
        """
        Retrieve members of the scope.

        .. versionchanged:: 3.7
           we added the supervisor members for backend that support this.

        :param is_manager: (otional) set to True/False to filter members that are/aren't managers, resp.
        :type is_manager: bool
        :param is_supervisor: (optional) set to True/False to filter members that are/aren't supervisors, resp.
        :type is_supervisor: bool
        :param is_leadmember: (optional) set to True/False to filter members that are/aren't leadmembers, resp.
        :type is_leadmember: bool
        :return: List of members, each defined as a dict

        Examples
        --------
        >>> members = project.members()
        >>> managers = project.members(is_manager=True)
        >>> supervisors = project.members(is_supervisor=True)
        >>> leadmembers = project.members(is_leadmember=True)

        """
        members = [
            member for member in self._json_data["members"] if member["is_active"]
        ]

        if is_manager is not None:
            members = [
                member for member in members if member.get("is_manager") == is_manager
            ]
        if is_supervisor is not None:
            members = [
                member
                for member in members
                if member.get("is_supervisor") == is_supervisor
            ]
        if is_leadmember is not None:
            members = [
                member
                for member in members
                if member.get("is_leadmember") == is_leadmember
            ]
        return members

    def add_member(self, member: Text) -> None:
        """
        Add a single member to the scope.

        You may only edit the list of members if the pykechain credentials allow this.

        :param member: single username to be added to the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.MEMBER, user=member
        )

    def remove_member(self, member: Text) -> None:
        """
        Remove a single member to the scope.

        :param member: single username to be removed from the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE, role=ScopeRoles.MEMBER, user=member
        )

    def add_manager(self, manager: Text) -> None:
        """
        Add a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.MANAGER, user=manager
        )

    def remove_manager(self, manager: Text) -> None:
        """
        Remove a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE, role=ScopeRoles.MANAGER, user=manager
        )

    def add_leadmember(self, leadmember: Text) -> None:
        """
        Add a single leadmember to the scope.

        :param leadmember: single username to be added to the scope list of leadmembers
        :type leadmember: basestring
        :raises APIError: when unable to update the scope leadmember
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.LEADMEMBER, user=leadmember
        )

    def remove_leadmember(self, leadmember: Text) -> None:
        """
        Remove a single leadmember to the scope.

        :param leadmember: single username to be added to the scope list of leadmembers
        :type leadmember: basestring
        :raises APIError: when unable to update the scope leadmember
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE,
            role=ScopeRoles.LEADMEMBER,
            user=leadmember,
        )

    def add_supervisor(self, supervisor: Text) -> None:
        """
        Add a single supervisor to the scope.

        .. versionadded:: 3.7
           requires backend version 3.7 as well.

        :param supervisor: single username to be added to the scope list of supervisors
        :type supervisor: basestring
        :raises APIError: when unable to update the scope supervisor
        """
        if self._client.match_app_version(label="scope", version="<3.6.0"):
            raise NotImplementedError(
                "Adding and removal of supervisor members to a scope not "
                "possible with this backend version"
            )
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.SUPERVISOR, user=supervisor
        )

    def remove_supervisor(self, supervisor: Text) -> None:
        """
        Remove a single supervisor to the scope.

        .. versionadded:: 3.7
           requires backend version 3.7 as well.

        :param supervisor: single username to be added to the scope list of supervisors
        :type supervisor: basestring
        :raises APIError: when unable to update the scope supervisor
        """
        if self._client.match_app_version(label="scope", version="<3.6.0"):
            raise NotImplementedError(
                "Adding and removal of supervisor members to a scope not "
                "possible with this backend version"
            )
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE,
            role=ScopeRoles.SUPERVISOR,
            user=supervisor,
        )



from pykechain.enums import SubprocessDisplayMode, KEChainPages, ScopeStatus
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models import Scope
from pykechain.models.sidebar.sidebar_button import SideBarButton
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from tests.classes import TestBetamax


class TestSideBar(TestBetamax):

    def setUp(self):
        super(TestSideBar, self).setUp()

        self.scope = self.project.clone(asynchronous=False)  # type: Scope
        del self.project

        self.manager = SideBarManager(scope=self.scope)
        self.scope_uri = '#/scopes/{}'.format(self.scope.id)
        self.scope_activity_uri = '{}/{}'.format(self.scope_uri, SubprocessDisplayMode.ACTIVITIES)

        self.default_button_config = {
            'title': 'new button',
            'icon': 'bookmark',
            'uri': '{}/{}'.format(self.scope_uri, KEChainPages.DATA_MODEL)
        }

    def tearDown(self):
        self.scope.delete()
        super(TestSideBar, self).tearDown()

    def test_manager(self):
        side_bar_manager = SideBarManager(scope=self.scope)

        self.assertIsInstance(side_bar_manager, SideBarManager)

    def test_singleton_manager_per_scope(self):
        manager_1 = SideBarManager(scope=self.scope)
        manager_2 = self.scope.side_bar()

        self.assertTrue(manager_1 is manager_2,
                        msg='There are 2 different manager objects while the same object is expected.')

    def test_loading_of_existing_buttons(self):
        for scope in self.client.scopes(status=ScopeStatus.ACTIVE):
            scope.side_bar()

    def test_override_sidebar_property(self):
        value = self.manager.override_sidebar

        self.assertIsInstance(value, bool)
        self.assertFalse(value)

        self.manager.override_sidebar = True

        self.assertTrue(self.manager.override_sidebar)

    def test_create_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        self.assertIsInstance(new_button, SideBarButton)
        self.assertTrue(new_button is self.manager[-1])
        self.assertIsInstance(new_button.__repr__(), str)

    # noinspection PyTypeChecker
    def test_create_button_incorrect_arguments(self):
        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order='3')

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title=False)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title='Button', icon=32)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title='Button', icon='pennant', uri=44)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title='Button', icon='pennant', uri='http://www.google.com',
                          uri_target='any')

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title='Button', icon='pennant', uri='http://www.google.com',
                          icon_mode='best')

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title='Button', icon='pennant', uri='http://www.google.com',
                          random='unsupported keyword')

    def test_edit_button(self):
        custom_name = "Custom Dutch name"
        new_button = self.manager.create_button(**self.default_button_config)

        new_button.edit(
            display_icon="pennant",
            displayName_nl=custom_name,
        )
        new_button.refresh()

        self.assertEqual("pennant", new_button.display_icon)
        self.assertEqual(custom_name, new_button._other_attributes.get("displayName_nl"))

    def test_delete_button(self):
        new_button = self.manager.create_button(**self.default_button_config)
        new_button.delete()

        self.assertNotIn(new_button, self.manager)

    def test_get_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        button_from_index = self.manager[0]
        button_from_title = self.manager['new button']
        button_from_button = self.manager[new_button]

        self.assertIs(new_button, button_from_index)
        self.assertIs(new_button, button_from_title)
        self.assertIs(new_button, button_from_button)

        with self.assertRaises(IndexError):
            # noinspection PyStatementEffect
            self.manager[3]

        with self.assertRaises(NotFoundError):
            # noinspection PyStatementEffect
            self.manager['unused button title']

    def test_remove_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        self.manager.remove(new_button)

        self.assertTrue(len(self.manager) == 0)

    def test_insert_button(self):
        button_1 = self.manager.create_button(order=0, **self.default_button_config)
        button_2 = self.manager.create_button(order=1, **self.default_button_config)
        button_3 = self.manager.create_button(order=2, **self.default_button_config)

        self.manager.insert(0, button_2)

        self.assertListEqual(
            [button_2, button_1, button_3],
            list(self.manager),
        )

    def test_bulk_add_buttons(self):
        buttons = [
            {
                'displayName': 'activity button',
                'displayIcon': 'bookmark',
                'uri': '{}/{}'.format(self.scope_activity_uri, self.scope.activities()[0].id),
            },
            {
                'displayName': 'ke-chain page',
                'displayIcon': 'university',
                'uri': '{}/{}'.format(self.scope_uri, KEChainPages.EXPLORER)
            },
            {
                'displayName': 'external button',
                'displayIcon': 'google',
                'uri': 'https://www.google.com',
            },
        ]
        new_buttons = self.manager.add_buttons(
            buttons=buttons,
            override_sidebar=True,
        )

        self.assertIsInstance(new_buttons, list)
        self.assertTrue(all(isinstance(b, SideBarButton) for b in new_buttons))
        self.assertTrue(len(self.manager) == 3)

    def test_context_manager(self):
        with SideBarManager(scope=self.scope) as bulk_creation_manager:

            bulk_creation_manager.override_sidebar = True

            bulk_creation_manager.add_task_button(
                activity=self.scope.activities()[0],
                icon='bookmark',
            )
            bulk_creation_manager.add_external_button(
                url='https://www.google.com',
                icon='google',
                title='Google',
            )
            bulk_creation_manager.add_ke_chain_page(
                page_name=KEChainPages.WORK_BREAKDOWN,
                icon='sitemap',
            )

            self.assertEqual(0, len(self.scope.options['customNavigation']),
                             msg='During bulk creation of buttons, KE-chain should not be updated yet.')

        self.assertEqual(3, len(self.scope.options['customNavigation']),
                         msg='After closing the context `with`, update should be performed.')

        updated_side_bar_buttons = self.scope.options.get('customNavigation')
        self.assertTrue(len(updated_side_bar_buttons) == 3,
                        msg='At the end of bulk creation, the buttons must have been created.')

        self.assertTrue(self.scope.options.get('overrideSideBar'))

    def test_load_buttons(self):
        # setup: create a custom button
        with self.scope.side_bar() as sbm:
            sbm.override_sidebar = True
            sbm.add_ke_chain_page(KEChainPages.WORK_BREAKDOWN)

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        # testing: delete original button (this tests the "load") and add another one
        with self.scope.side_bar() as sbm:
            for button in sbm:
                sbm.delete_button(button)
            sbm.override_sidebar = True
            sbm.add_ke_chain_page(KEChainPages.TASKS)

        self.assertEqual(1, len(self.scope.options.get('customNavigation')))

    def test_attributes_button(self):
        display_name_nl = "Het data model"
        button_title = "New button"

        # Check whether all attributes are retrieved and set
        with self.scope.side_bar() as sbm:
            sbm.add_ke_chain_page(
                page_name=KEChainPages.DATA_MODEL,
                title=button_title,
                displayName_nl=display_name_nl,
            )

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        sbm = self.scope.side_bar()
        found = False
        for button in sbm:  # type: SideBarButton
            if button._other_attributes.get("displayName_nl") == display_name_nl:
                found = True

        self.assertTrue(found)

    def test_attributes_sidebar(self):
        sbm = self.scope.side_bar()

        # Check starting condition
        self.assertFalse(sbm.override_sidebar)

        # Set value
        sbm.override_sidebar = True

        # Reload side-bar from KE-chain
        sbm.refresh()

        # Test attributes
        self.assertTrue(sbm.override_sidebar)



DESCRIPTION = """pykechain is a an open source Python API to KE-chain. Provide remote access and control through the
public KE-chain API. Works only in combination with a valid KE-chain license. KE-chain BV (c) 2016 """

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

PACKAGE_NAME = "pykechain"
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

ABOUT = {}
with open(path.join(HERE, PACKAGE_NAME, "__about__.py")) as f:
    exec(f.read(), ABOUT)

setup(
    name=ABOUT["name"],
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=ABOUT["version"],
    description=ABOUT["description"],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # The project's main homepage.
    url="https://github.com/KE-works/pykechain",
    # Author details
    author=ABOUT["author"],
    author_email=ABOUT["email"],
    # Choose your license
    license="Apache Open Source License 2.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
        # OS
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
    ],
    # What does your project relate to?
    keywords="python api rest sdk KE-chain",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["tests"]),
    # Project URLs
    project_urls={
        "Documentation": "https://pykechain.readthedocs.io/en/latest",
        "Changelog": "https://github.com/KE-works/pykechain/blob/master/CHANGELOG.rst",
        "Source": "https://github.com/KE-works/pykechain/",
        "Tracker": "https://github.com/KE-works/pykechain/issues",
        "Company Page": "https://ke-chain.com",
    },
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "requests>=2.20.0",
        "envparse",
        "typing",
        "six",
        "jsonschema",
        "semver>=2.10.0",
        "pytz",
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "betamax"],
    python_requires=">=3.5",
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
)
